import subprocess
import logging
import hydra
from ekorpkit import eKonf, __hydra_version_base__


logger = eKonf.getLogger(__name__)


def run_job(**args):
    command = args.get("ekorpkit") or {}
    command = [f"{k}={v}" for k, v in command.items()]
    logger.info(f"Running ekorpkit with args: {command}")
    _run_cmd(command)


def _run_cmd(command):
    command = ["ekorpkit"] + command
    proc = subprocess.Popen(
        " ".join(command),
        shell=True,
    )
    proc.wait()


def run_workflow(**args):
    _workflows = args.get("workflow") or {}
    _default_args = args.get("ekorpkit") or {}
    _default_args = [f"{k}={v}" for k, v in _default_args.items()]
    for wf_name, workflow in _workflows.items():
        logger.info("Running workflow: {}".format(wf_name))
        _jobs = workflow.get("jobs") or []
        for job in _jobs:
            job_arg = _default_args + [f"+job/{wf_name}={job}"]
            logger.info(f"Running job [{job}] with args: {job_arg}")
            _run_cmd(job_arg)


@hydra.main(config_path="conf", config_name="run", version_base=__hydra_version_base__)
def main(cfg) -> None:
    verbose = cfg.verbose

    if verbose:
        logger.info("## eKorpkit-flow Command Line Interface##")
    eKonf.load_dotenv(verbose)

    if cfg.get("verbose"):
        logger.info("## hydra configuration resolved ##")
        eKonf.pprint(cfg)

    eKonf.instantiate(cfg)


if __name__ == "__main__":
    main()



import pandas as pd
import logging
import os
import sklearn
from abc import ABCMeta, abstractmethod
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class AutoML:
    __metaclass__ = ABCMeta
    Keys = eKonf.Keys

    def __init__(self, **args):
        from flaml import AutoML

        self._dataset_cfg = args.pop(eKonf.Keys.DATASET, None)

        args = eKonf.to_config(args)
        self.args = args
        self.name = args.name
        self.verbose = args.get("verbose", True)
        self._model_cfg = eKonf.to_dict(args.config)
        self._columns = args.get(eKonf.Keys.COLUMNS)
        self._train_ = args[eKonf.Keys.TRAIN]
        self._predict_ = args[eKonf.Keys.PREDICT]
        self._method_ = self.args.get(eKonf.Keys.METHOD)
        self._eval_cfg = args.model.eval
        self._learning_curve = args.learning_curve.visualize.plot
        self._feature_importance = args.feature_importance.visualize.plot

        self._path = self.args.path
        self._model_file = self._path.model.filepath
        self._log_file = self._path.log.filepath
        self._pred_file = self._path.pred.filepath

        self._automl = AutoML()
        self._dataset = None
        self._X_train = None
        self._X_dev = None
        self._X_test = None
        self._y_train = None
        self._y_dev = None
        self._y_test = None
        self._classes = None
        self.pred_data = None

        eKonf.methods(self._method_, self)

    def fit(self):
        if self.X_train is None:
            self.load_dataset()

        X_train = self.X_train.values
        y_train = self.dataset.transform_labels(self.y_train.values)
        log.info(f"types of X_train: {type(X_train)}, y_train: {type(y_train)}")
        log.info(f"fitting a model with {self._model_cfg}")
        self._automl.fit(X_train=X_train, y_train=y_train, **self._model_cfg)
        # Print the results
        self.show_results()

    def save(self):
        """pickle and save the automl object"""
        import pickle

        with open(self._model_file, "wb") as f:
            pickle.dump(self._automl, f, pickle.HIGHEST_PROTOCOL)
        log.info(f"Saved model to {self._model_file}")

    def load(self):
        import pickle

        with open(self._model_file, "rb") as f:
            self._automl = pickle.load(f)
        log.info(f"Loaded model from {self._model_file}")

    @property
    def dataset(self):
        return self._dataset

    @property
    def classes(self):
        return self._classes

    @property
    def best_estimator(self):
        return self._automl.model.estimator

    def show_results(self):
        """retrieve best config and best learner"""
        print("Best ML leaner:", self._automl.best_estimator)
        print("Best hyperparmeter config:", self._automl.best_config)
        print(
            "Best accuracy on validation data: {0:.4g}".format(
                1 - self._automl.best_loss
            )
        )
        print(
            "Training duration of best run: {0:.4g} s".format(
                self._automl.best_config_train_time
            )
        )

    def get_logs(self, time_budget=240):
        from flaml.data import get_output_from_log

        if not eKonf.exists(self._log_file):
            log.warning(f"Log file {self._log_file} not found")
            return None
        (
            time_history,
            best_valid_loss_history,
            valid_loss_history,
            config_history,
            metric_history,
        ) = get_output_from_log(filename=self._log_file, time_budget=time_budget)
        return {
            "time_history": time_history,
            "best_valid_loss_history": best_valid_loss_history,
            "valid_loss_history": valid_loss_history,
            "config_history": config_history,
            "metric_history": metric_history,
        }

    def _predict(self, X):
        """compute predictions of testing dataset"""
        y_pred = self._automl.predict(X)
        y_pred = self.dataset.inverse_transform_labels(y_pred)
        return {self.Keys.PREDICTED: y_pred}

    def convert_to_X(self, data):
        X_cols = self._predict_[self.Keys.X]
        if X_cols is None:
            X_cols = list(data.columns)
        X = data[X_cols].values
        if self.verbose:
            print(X[:5])
        return X

    def append_predictions(self, data, preds):
        predicted_column = self._predict_[self.Keys.PREDICTED]
        data = data.copy()
        data[predicted_column] = preds[self.Keys.PREDICTED]
        return data

    def predict(self, data, _predict_={}):
        if _predict_:
            self._predict_ = _predict_
        X = self.convert_to_X(data)
        preds = self._predict(X)
        return self.append_predictions(data, preds)

    def eval(self):
        from flaml.ml import sklearn_metric_loss_score

        if self.X_test is None:
            self.load_dataset()

        if self.X_test is None:
            log.warning("No test data found")
            return

        y_preds = self._predict(self.X_test.values)
        pred_data = self.append_predictions(self.X_test, y_preds)
        actual_column = self._eval_cfg._eval_[self.Keys.ACTUAL]
        if actual_column:
            pred_data[actual_column] = self.y_test
        eKonf.save_data(pred_data, self._pred_file)
        if self.verbose:
            print(pred_data.head())

        y_pred = self.dataset.transform_labels(y_preds[self.Keys.PREDICTED])
        y_test = self.dataset.transform_labels(self.y_test)
        print("r2:", 1 - sklearn_metric_loss_score("r2", y_pred, y_test))
        print("mse:", sklearn_metric_loss_score("mse", y_pred, y_test))
        print("mae:", sklearn_metric_loss_score("mae", y_pred, y_test))

        self._eval_cfg.labels = self.classes
        eKonf.instantiate(self._eval_cfg, data=pred_data)
        self.pred_data = pred_data

    def load_dataset(self):
        if self._dataset_cfg is None:
            log.warning("No dataset config found")
            return
        if eKonf.is_instantiatable(self._dataset_cfg):
            self._dataset = eKonf.instantiate(self._dataset_cfg)
        else:
            self._dataset = self._dataset_cfg
        self._dataset.fit_labelencoder(self._dataset.y_train)
        self._classes = self._dataset.classes

        self._X_train = self._dataset.X_train
        self._X_dev = self._dataset.X_dev
        self._X_test = self._dataset.X_test
        self._y_train = self._dataset.y_train
        self._y_dev = self._dataset.y_dev
        self._y_test = self._dataset.y_test

        if self.verbose:
            print("Train data:")
            print(self._X_train.tail())
            if self._X_dev is not None:
                print("Eval data:")
                print(self._X_dev.tail())
            else:
                log.info("No eval data found")
            if self._X_test is not None:
                print("Test data:")
                print(self._X_test.tail())
            else:
                log.info("No test data found")

    @property
    def X_train(self):
        return self._X_train

    @property
    def X_dev(self):
        return self._X_dev

    @property
    def X_test(self):
        return self._X_test

    @property
    def y_train(self):
        return self._y_train

    @property
    def y_dev(self):
        return self._y_dev

    @property
    def y_test(self):
        return self._y_test

    def get_feature_importance(self, estimator=None, n_features=None):
        if estimator is None:
            estimator = self.best_estimator
        if self.X_train is None:
            self.load_dataset()
        _data = {
            "columns": self.X_train.columns.tolist(),
            "importances": estimator.feature_importances_.tolist(),
        }
        data = pd.DataFrame(_data)
        data.sort_values(by="importances", ascending=False, inplace=True)
        if n_features is not None:
            data = data.head(n_features)

        return data

    def plot_feature_importance(self, estimator=None, n_features=None):
        data = self.get_feature_importance(estimator=estimator, n_features=n_features)

        eKonf.instantiate(self._feature_importance, data=data)

    def get_log_data(self):
        train_logs = self.get_logs()

        data = pd.DataFrame(train_logs)
        data["acc_history"] = 1 - data["valid_loss_history"]
        data["best_acc_history"] = 1 - data["best_valid_loss_history"]
        return data

    def plot_learning_curve(self):
        data = self.get_log_data()

        eKonf.instantiate(self._learning_curve, data=data)



import sys
import re
import requests
import orjson as json
from bs4 import BeautifulSoup
from datetime import datetime

# Import parent class
from .base import FOMC


class Testimony(FOMC):
    """
    A convenient class for extracting testimony from the FOMC website.
    Among testimonies, there are semi annual monetary policy report to the Congress by chairperson.
    """

    def __init__(self, content_type, **args):
        super().__init__(content_type, **args)

    def _get_links(self, from_year):
        """
        Override private function that sets all the links for the contents to download on FOMC website
         from from_year (=min(1996, from_year)) to the current most recent year
        """
        self.links = []
        self.titles = []
        self.speakers = []
        self.dates = []

        if self.verbose:
            print("Getting links for testimony...")
        to_year = datetime.today().strftime("%Y")

        if from_year < 1996:
            print("Archive only from 1996, so setting from_year as 1996...")
            from_year = 1996
        elif from_year > 2006:
            print(
                "All data from 2006 is in a single json, so return all from 2006 anyway though specified from year is ",
                from_year,
            )

        url = self.base_url + "/json/ne-testimony.json"
        res = requests.get(url)
        res_list = json.loads(res.content.decode("utf-8-sig"))
        for record in res_list:
            doc_link = record.get("l")
            if doc_link:
                self.links.append(doc_link)
                self.titles.append(record.get("t"))
                self.speakers.append(record.get("s"))
                date_str = record.get("d").split(" ")[0]
                self.dates.append(datetime.strptime(date_str, "%m/%d/%Y"))

        if from_year < 2006:
            for year in range(from_year, 2006):
                url = (
                    self.base_url
                    + "/newsevents/testimony/"
                    + str(year)
                    + "testimony.htm"
                )

                res = requests.get(url)
                soup = BeautifulSoup(res.text, "html.parser")

                doc_links = soup.findAll(
                    "a",
                    href=re.compile(
                        "^/boarddocs/testimony/{}/|^/boarddocs/hh/{}/".format(
                            str(year), str(year)
                        )
                    ),
                )
                for doc_link in doc_links:
                    # Sometimes the same link is put for watch live video. Skip those.
                    if doc_link.find({"class": "watchLive"}):
                        continue
                    # Add links
                    self.links.append(doc_link.attrs["href"])

                    # Handle mark-up mistakes
                    if doc_link.get("href") in (
                        "/boarddocs/testimony/2005/20050420/default.htm"
                    ):
                        title = doc_link.get_text()
                        speaker = (
                            doc_link.parent.parent.next_element.next_element.get_text()
                            .replace("\n", "")
                            .strip()
                        )
                        date_str = doc_link.parent.parent.next_element.replace(
                            "\n", ""
                        ).strip()
                    elif doc_link.get("href") in (
                        "/boarddocs/testimony/1997/19970121.htm"
                    ):
                        title = (
                            doc_link.parent.parent.find_next("em")
                            .get_text()
                            .replace("\n", "")
                            .strip()
                        )
                        speaker = (
                            doc_link.parent.parent.find_next("strong")
                            .get_text()
                            .replace("\n", "")
                            .strip()
                        )
                        date_str = doc_link.get_text()
                    else:
                        title = doc_link.get_text()
                        speaker = (
                            doc_link.parent.find_next("div")
                            .get_text()
                            .replace("\n", "")
                            .strip()
                        )
                        # When a video icon is placed between the link and speaker
                        if speaker in ("Watch Live", "Video"):
                            speaker = (
                                doc_link.parent.find_next("p")
                                .find_next("p")
                                .get_text()
                                .replace("\n", "")
                                .strip()
                            )
                        date_str = doc_link.parent.parent.next_element.replace(
                            "\n", ""
                        ).strip()

                    self.titles.append(doc_link.get_text())
                    self.speakers.append(speaker)
                    self.dates.append(datetime.strptime(date_str, "%B %d, %Y"))

                if self.verbose:
                    print(
                        "YEAR: {} - {} testimony docs found.".format(
                            year, len(doc_links)
                        )
                    )

    def _add_article(self, link, index=None):
        """
        Override a private function that adds a related article for 1 link into the instance variable
        The index is the index in the article to add to.
        Due to concurrent prcessing, we need to make sure the articles are stored in the right order
        """
        if self.verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        link_url = self.base_url + link
        # article_date = self._date_from_link(link)

        # print(link_url)

        # date of the article content
        # self.dates.append(article_date)

        res = requests.get(self.base_url + link)
        html = res.text
        # p tag is not properly closed in many cases
        html = html.replace("<P", "<p").replace("</P>", "</p>")
        html = html.replace("<p", "</p><p").replace("</p><p", "<p", 1)
        # remove all after appendix or references
        x = re.search(
            r"(<b>references|<b>appendix|<strong>references|<strong>appendix)",
            html.lower(),
        )
        if x:
            html = html[: x.start()]
            html += "</body></html>"
        # Parse html text by BeautifulSoup
        article = BeautifulSoup(html, "html.parser")
        # Remove footnote
        for fn in article.find_all("a", {"name": re.compile("fn\d")}):
            # if fn.parent:
            #     fn.parent.decompose()
            # else:
            fn.decompose()
        # Get all p tag
        paragraphs = article.findAll("p")
        para_text = []
        for paragraph in paragraphs:
            text = re.sub(r"\s+", " ", paragraph.get_text().strip())
            if len(text) > 0 and text not in self.skip_text_list:
                para_text.append(text)
        self.articles[index] = self.segment_separator.join(para_text)



# original source from https://github.com/yuki678/centralbank_analysis
from .statement import Statement
from .minutes import Minutes
from .meeting import MeetingScript
from .press_conf import PresConfScript
from .speech import Speech
from .testimony import Testimony
from .beigebook import BeigeBook
from .base import build_fomc, FOMC



from typing import Tuple

from ..types import Config


class SGDOptimizerConfig(Config):
    """Settings for SGD optimizer."""

    momentum: float = 0.9


class AdamOptimizerConfig(Config):
    """Settings for Adam optimizer."""

    amsgrad: bool = False
    betas: Tuple[float, float] = (0.9, 0.999)


class AdamaxOptimizerConfig(Config):
    """Settings for Adamax optimizer."""

    betas: Tuple[float, float] = (0.9, 0.999)
    eps: float = 1e-8


class OptimizerConfig(Config):
    """Settings common to all optimizers."""

    sgd_config: SGDOptimizerConfig = SGDOptimizerConfig()  # type:ignore
    adam_config: AdamOptimizerConfig = AdamOptimizerConfig()  # type:ignore
    adamax_config: AdamaxOptimizerConfig = AdamaxOptimizerConfig()  # type:ignore



from typing import Tuple

import numpy as np
import pandas as pd


def filter_unlabeled_dataframe(
    X: pd.DataFrame, y: np.ndarray, L: np.ndarray
) -> Tuple[pd.DataFrame, np.ndarray]:
    """Filter out examples not covered by any labeling function.

    Parameters
    ----------
    X
        Data points in a Pandas DataFrame.
    y
        Matrix of probabilities output by label model's predict_proba method.
    L
        Matrix of labels emitted by LFs.

    Returns
    -------
    pd.DataFrame
        Data points that were labeled by at least one LF in L.
    np.ndarray
        Probabilities matrix for data points labeled by at least one LF in L.
    """
    mask = (L != -1).any(axis=1)
    return X.iloc[mask], y[mask]



from ...labeling import LFApplier, PandasLFApplier


class SFApplier(LFApplier):
    """SF applier for a list of data points.

    See ``snorkel.labeling.core.LFApplier`` for details.
    """

    _use_recarray = True


class PandasSFApplier(PandasLFApplier):
    """SF applier for a Pandas DataFrame.

    See ``snorkel.labeling.core.PandasLFApplier`` for details.
    """

    _use_recarray = True



import os
import logging
import hashlib
import shutil
import pandas as pd
import numpy as np
from pathlib import Path
from ekorpkit.io.fetch.web import web_download, web_download_unzip


log = logging.getLogger(__name__)


def _download_models(
    name,
    path,
    archive_path,
    link,
    link_fb=None,
    SHA=None,
    check_model_SHA=False,
    zip_path=None,
    unzip=False,
    **kwargs,
):

    if os.path.exists(path) and check_model_SHA and SHA:
        log.info(f"Checking {name} File")
        with open(path, "rb") as f:
            bytes = f.read()
            hash = hashlib.sha256(bytes).hexdigest()
        if hash == SHA:
            log.info(f"Model {name} SHA matches")
        else:
            log.info(f"Model {name} SHA doesn't match, redownloading...")
            os.remove(path)
            _download_models(
                name, path, archive_path, link, link_fb, SHA, check_model_SHA
            )
    elif os.path.exists(path):
        log.info(
            f"Model {name} already downloaded, set check_model_SHA to true if the file is corrupt"
        )
    elif os.path.exists(archive_path) and check_model_SHA and SHA:
        log.info(f"Checking {name} File")
        with open(archive_path, "rb") as f:
            bytes = f.read()
            hash = hashlib.sha256(bytes).hexdigest()
        if hash == SHA:
            log.info(f"Model {name} SHA matches")
            log.info(f"Copying {name} File to {path} from {archive_path}")
            shutil.copyfile(archive_path, path)
        else:
            log.info(f"Model {name} SHA doesn't match, redownloading...")
            os.remove(archive_path)
            _download_models(name, path, archive_path, link, link_fb)
    elif os.path.exists(archive_path):
        log.info(f"Copying {name} File to {path} from {archive_path}")
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(archive_path, path)
    else:
        if unzip:
            web_download_unzip(link, zip_path)
        else:
            web_download(link, archive_path)
        if os.path.exists(archive_path):
            log.info(f"Copying {name} File to {path} from {archive_path}")
            Path(path).parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(archive_path, path)
        elif not os.path.exists(archive_path) and link_fb:
            log.info("First URL Failed using FallBack")
            _download_models(name, path, archive_path, link_fb)
        else:
            log.warning(f"{name} File not found")


def move_files(start_num, end_num, old_folder, new_folder, batch_name, batch_num):
    for i in range(start_num, end_num):
        old_file = old_folder + f"/{batch_name}({batch_num})_{i:04}.png"
        new_file = new_folder + f"/{batch_name}({batch_num})_{i:04}.png"
        os.rename(old_file, new_file)


def split_prompts(prompts, max_frames):
    prompt_series = pd.Series([np.nan for a in range(max_frames)])
    for i, prompt in prompts.items():
        if isinstance(prompt, str):
            prompt =[prompt]
        prompt_series[i] = prompt
    # prompt_series = prompt_series.astype(str)
    prompt_series = prompt_series.ffill().bfill()
    return prompt_series


def parse_key_frames(string, prompt_parser=None):
    """Given a string representing frame numbers paired with parameter values at that frame,
    return a dictionary with the frame numbers as keys and the parameter values as the values.

    Parameters
    ----------
    string: string
        Frame numbers paired with parameter values at that frame number, in the format
        'framenumber1: (parametervalues1), framenumber2: (parametervalues2), ...'
    prompt_parser: function or None, optional
        If provided, prompt_parser will be applied to each string of parameter values.

    Returns
    -------
    dict
        Frame numbers as keys, parameter values at that frame number as values

    Raises
    ------
    RuntimeError
        If the input string does not match the expected format.

    Examples
    --------
    >>> parse_key_frames("10:(Apple: 1| Orange: 0), 20: (Apple: 0| Orange: 1| Peach: 1)")
    {10: 'Apple: 1| Orange: 0', 20: 'Apple: 0| Orange: 1| Peach: 1'}

    >>> parse_key_frames("10:(Apple: 1| Orange: 0), 20: (Apple: 0| Orange: 1| Peach: 1)", prompt_parser=lambda x: x.lower()))
    {10: 'apple: 1| orange: 0', 20: 'apple: 0| orange: 1| peach: 1'}
    """
    import re

    pattern = r"((?P<frame>[0-9]+):[\s]*[\(](?P<param>[\S\s]*?)[\)])"
    frames = dict()
    for match_object in re.finditer(pattern, string):
        frame = int(match_object.groupdict()["frame"])
        param = match_object.groupdict()["param"]
        if prompt_parser:
            frames[frame] = prompt_parser(param)
        else:
            frames[frame] = param

    if frames == {} and len(string) != 0:
        raise RuntimeError("Key Frame string not correctly formatted")
    return frames


def get_inbetweens(key_frames, max_frames, interp_spline, integer=False):
    """Given a dict with frame numbers as keys and a parameter value as values,
    return a pandas Series containing the value of the parameter at every frame from 0 to max_frames.
    Any values not provided in the input dict are calculated by linear interpolation between
    the values of the previous and next provided frames. If there is no previous provided frame, then
    the value is equal to the value of the next provided frame, or if there is no next provided frame,
    then the value is equal to the value of the previous provided frame. If no frames are provided,
    all frame values are NaN.

    Parameters
    ----------
    key_frames: dict
        A dict with integer frame numbers as keys and numerical values of a particular parameter as values.
    integer: Bool, optional
        If True, the values of the output series are converted to integers.
        Otherwise, the values are floats.

    Returns
    -------
    pd.Series
        A Series with length max_frames representing the parameter values for each frame.

    Examples
    --------
    >>> max_frames = 5
    >>> get_inbetweens({1: 5, 3: 6})
    0    5.0
    1    5.0
    2    5.5
    3    6.0
    4    6.0
    dtype: float64

    >>> get_inbetweens({1: 5, 3: 6}, integer=True)
    0    5
    1    5
    2    5
    3    6
    4    6
    dtype: int64
    """
    key_frame_series = pd.Series([np.nan for a in range(max_frames)])

    for i, value in key_frames.items():
        key_frame_series[i] = value
    key_frame_series = key_frame_series.astype(float)

    interp_method = interp_spline

    if interp_method == "Cubic" and len(key_frames.items()) <= 3:
        interp_method = "Quadratic"

    if interp_method == "Quadratic" and len(key_frames.items()) <= 2:
        interp_method = "Linear"

    key_frame_series[0] = key_frame_series[key_frame_series.first_valid_index()]
    key_frame_series[max_frames - 1] = key_frame_series[
        key_frame_series.last_valid_index()
    ]
    # key_frame_series = key_frame_series.interpolate(method=intrp_method,order=1, limit_direction='both')
    key_frame_series = key_frame_series.interpolate(
        method=interp_method.lower(), limit_direction="both"
    )
    if integer:
        return key_frame_series.astype(int)
    return key_frame_series



import os
import logging
import re
import math
import random
import numpy as np
import gc
import matplotlib.pyplot as plt
import shutil
import torch
import torchvision.transforms as T
import torchvision.transforms.functional as TF
import PIL
from PIL import Image
from tqdm.auto import tqdm
from glob import glob
from pathlib import Path
from ekorpkit import eKonf
from .utils import (
    _download_models,
    move_files,
    split_prompts,
    parse_key_frames,
    get_inbetweens,
)
from ekorpkit.utils.func import elapsed_timer
from ekorpkit.visualize.motion import create_video as _create_video
from ekorpkit.visualize.motion import extract_frames
from ..art.base import BaseTTIModel
from enum import Enum

log = logging.getLogger(__name__)


class AnimMode(str, Enum):
    """Animation mode"""

    NONE = "None"
    ANIM_2D = "2D"
    ANIM_3D = "3D"
    VIDEO_INPUT = "Video Input"


class DiscoDiffusion(BaseTTIModel):

    TRANSLATION_SCALE = 1.0 / 200.0

    def __init__(self, **args):
        super().__init__(**args)

        self._parameters = self.args.parameters
        self.midas_config = self.args.midas
        # self.model_config = self.args.model
        self.model_map = self.args.model_map
        self.model_map.diffusion_models_256x256_list += (
            self.model_map.kaliyuga_pixel_art_model_names
            + self.model_map.kaliyuga_watercolor_model_names
            + self.model_map.kaliyuga_pulpscifi_model_names
        )
        self.diffusion_model_config = self.args.diffusion_model
        self._download = self.args.download

        self._model_and_diffusion_args = {}
        self.clip_models = []
        self.model = None
        self.diffusion = None
        self.angle_series = None
        self.zoom_series = None
        self.translation_x_series = None
        self.translation_y_series = None
        self.translation_z_series = None
        self.rotation_3d_x_series = None
        self.rotation_3d_y_series = None
        self.rotation_3d_z_series = None
        self.text_prompts_series = None
        self.image_prompts_series = None

        if self.auto.load:
            self.load()

    def batch_imagine(
        self,
        text_prompts=None,
        image_prompts=None,
        batch_name=None,
        batch_args=None,
        batch_pairs=None,
        show_collage=True,
        **args,
    ):
        """Run a batch"""
        animation_mode = AnimMode.NONE
        n_samples = 1
        args.update(dict(show_collage=False))

        if batch_args is None:
            batch_args = {}
        if not isinstance(batch_args, dict):
            raise ValueError("batch_args must be a dictionary")
        if len(batch_args) < 1:
            raise ValueError("batch_args must have at least 1 element")
        for k, v in batch_args.items():
            if not isinstance(v, (list, tuple)):
                v = [v]
        if batch_pairs is None:
            batch_pairs = [[arg_name] for arg_name in batch_args.keys()]

        _batch_results = []
        for batch_pair in batch_pairs:
            batch_pair_args = {}
            for arg_name in batch_pair:
                batch_pair_args[arg_name] = batch_args[arg_name]
            _batch_name = batch_name + "_" + "_".join(batch_pair)
            batch_config = dict(
                batch_name=_batch_name,
                batch_pair_args=batch_pair_args,
                text_prompts=text_prompts,
                image_prompts=image_prompts,
                results=[],
            )
            for pair_args in eKonf.dict_product(batch_pair_args):
                _args = args.copy()
                _args.update(pair_args)
                print(f"batch: {_batch_name} with {pair_args}")
                results = self.imagine(
                    text_prompts=text_prompts,
                    image_prompts=image_prompts,
                    batch_name=_batch_name,
                    animation_mode=animation_mode,
                    n_samples=n_samples,
                    **_args,
                )
                batch_config["results"].append(
                    dict(
                        args=pair_args,
                        config_file=results["config_file"],
                        image_filepaths=results["image_filepaths"],
                    )
                )

            _batch_config_path = self.save_batch_configs(batch_config)
            _batch_results.append(_batch_config_path)
            if show_collage:
                self.batch_collage(_batch_config_path)
        return _batch_results

    def save_batch_configs(self, args):
        """Save the settings"""
        _batch_name = args["batch_name"]
        _filename = f"{_batch_name}_batch_configs.yaml"
        _path = os.path.join(self._output.batch_configs_dir, _filename)
        log.info(f"Saving batch configs to {_path}")
        eKonf.save(args, _path)
        return _path

    def batch_collage(
        self,
        batch_config_path,
        xlabel=None,
        ylabel=None,
        zlabel=None,
        prompt_fontsize=18,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=18,
        fontcolor=None,
        **kwargs,
    ):
        args = eKonf.load(batch_config_path)
        args = eKonf.to_dict(args)

        batch_name = args["batch_name"]
        batch_pair_args = args["batch_pair_args"]
        arg_names = list(batch_pair_args.keys())

        if ylabel is None:
            ylabel = arg_names[0]
        if ylabel in arg_names:
            yticklabels = batch_pair_args[ylabel]
            arg_names.remove(ylabel)
        else:
            raise ValueError(f"{ylabel} not in {arg_names}")
        # reverse yticklabels so that the first image is the top left
        yticklabels = yticklabels[::-1]

        if xlabel is None and len(arg_names) > 0:
            xlabel = arg_names[0]
        if xlabel in arg_names:
            xticklabels = batch_pair_args[xlabel]
            arg_names.remove(xlabel)
        else:
            xticklabels = None

        if zlabel is None and len(arg_names) > 0:
            zlabel = arg_names[0]
        if zlabel in arg_names:
            ztitles = batch_pair_args[zlabel]
            arg_names.remove(zlabel)
        else:
            ztitles = [None]

        results = args["results"]
        prompt = args["text_prompts"]
        ncols = 1 if xticklabels is None else len(xticklabels)

        log.info(f"Prompt: {prompt}")
        for ztitle in ztitles:
            title = f"batch name: {batch_name}\nprompt: {prompt}\n"
            if ztitle is not None:
                output_filepath = batch_config_path.replace(
                    ".yaml", f"_{zlabel}({ztitle}).png"
                )
                title += f"{zlabel}: {ztitle}\n"
            else:
                output_filepath = batch_config_path.replace(".yaml", ".png")

            image_filepaths = []
            for result in results:
                if zlabel is None or result["args"][zlabel] == ztitle:
                    image_filepaths.append(result["image_filepaths"][0])

            eKonf.collage(
                image_filepaths=image_filepaths,
                output_filepath=output_filepath,
                ncols=ncols,
                title=title,
                title_fontsize=prompt_fontsize,
                show_filename=show_filename,
                filename_offset=filename_offset,
                fontname=fontname,
                fontsize=fontsize,
                fontcolor=fontcolor,
                xlabel=xlabel,
                ylabel=ylabel,
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                xlabel_fontsize=fontsize,
                ylabel_fontsize=fontsize,
                **kwargs,
            )
            print(f"Saved collage to {output_filepath}")

    def imagine_generator(
        self,
        text_prompts=None,
        image_prompts=None,
        batch_name=None,
        batch_num=None,
        **args,
    ):
        animation_mode = AnimMode.NONE
        args.update(dict(animation_mode=animation_mode))

        if text_prompts is not None:
            args.update(dict(text_prompts=text_prompts))
        if image_prompts is not None:
            args.update(dict(image_prompts=image_prompts))

        log.info("> loading settings...")
        args = self.load_config(batch_name=batch_name, batch_num=batch_num, **args)
        args = self._prepare_config(args)
        self.save_settings(args)
        self._config = args

        self._prepare_models()
        self.sample_imagepaths = []

        gc.collect()
        torch.cuda.empty_cache()
        try:
            for sample in self._generator(args):
                yield sample
        finally:
            gc.collect()
            torch.cuda.empty_cache()

        res = {"config": eKonf.to_dict(args), "imagepaths": self.sample_imagepaths}
        yield res

    def save_init_image(self, batch_name, image):
        self.load_config(batch_name)
        batch_dir = self._output.batch_dir
        init_image_path = os.path.join(batch_dir, f"{batch_name}_init.png")
        image.save(init_image_path)
        return init_image_path

    def imagine(
        self,
        text_prompts=None,
        image_prompts=None,
        batch_name=None,
        batch_num=None,
        animation_mode: AnimMode = None,
        run_to_resume=None,
        show_collage=True,
        show_losses=False,
        **args,
    ):

        if text_prompts is not None:
            args.update(dict(text_prompts=text_prompts))
        if image_prompts is not None:
            args.update(dict(image_prompts=image_prompts))
        if animation_mode is not None:
            args.update(dict(animation_mode=animation_mode))
        if run_to_resume is not None:
            args.update(dict(run_to_resume=run_to_resume))
        if show_collage is not None:
            args.update(dict(show_collage=show_collage))
        if show_losses is not None:
            args.update(dict(show_losses=show_losses))

        log.info("> loading settings...")
        args = self.load_config(batch_name=batch_name, batch_num=batch_num, **args)
        args = self._prepare_config(args)
        self._config = args

        if args.animation_mode == AnimMode.NONE:
            if args.start_sample >= args.n_samples:
                log.warning(
                    f"start_sample ({args.start_sample}) must be less than n_samples ({args.n_samples})"
                )
                return
        else:
            if args.start_frame >= args.max_frames:
                log.warning(
                    f"start frame {args.start_frame} is greater than max frames {args.max_frames}"
                )
                return

        log.info(
            f"Starting Run: {args.batch_name}({args.batch_num}) at frame {args.start_frame}"
        )
        self._prepare_models()
        self.sample_imagepaths = []
        with elapsed_timer(format_time=True) as elapsed:

            gc.collect()
            torch.cuda.empty_cache()
            loss_values = []
            try:
                if args.animation_mode == AnimMode.NONE:
                    loss_values = self._run(args)
                else:
                    self._run_anim(args)
            except KeyboardInterrupt:
                pass
            finally:
                log.info(f"Seed used: {args.seed}")
                gc.collect()
                torch.cuda.empty_cache()

            log.info(" >> elapsed time to diffuse: {}".format(elapsed()))
            if args.animation_mode == AnimMode.NONE:
                print(f"{args.n_samples} samples generated to {self._output.batch_dir}")
                print(f"text prompts: {text_prompts}")
                print("sample image paths:")
                for p in self.sample_imagepaths:
                    print(p)

                if args.show_collage:
                    eKonf.clear_output()
                    self.collage(image_filepaths=self.sample_imagepaths)

        results = {
            "image_filepaths": self.sample_imagepaths,
            "config_file": self.save_settings(args),
            "config": eKonf.to_dict(args),
            "loss_values": loss_values,
        }
        return results

    def get_model_path(self, model_name):
        model_filepath = self._download.models[model_name]["path"]
        return model_filepath

    def _prepare_models(self):
        from guided_diffusion.script_util import create_model_and_diffusion

        log.info("Prepping model...")
        self.model, self.diffusion = create_model_and_diffusion(
            **self._model_and_diffusion_args
        )
        self.diffusion_model = self.config.models.diffusion_model
        if self.diffusion_model == "custom":
            custom_path = os.path.join(
                self._path.model_dir, self.config.models.custom_model
            )
            self.model.load_state_dict(torch.load(custom_path, map_location="cpu"))
        else:
            model_path = self.get_model_path(self.diffusion_model)
            self.model.load_state_dict(torch.load(model_path, map_location="cpu"))
        self.model.requires_grad_(False).eval().to(self.cuda_device)
        for name, param in self.model.named_parameters():
            if "qkv" in name or "norm" in name or "proj" in name:
                param.requires_grad_()
        if self._model_and_diffusion_args["use_fp16"]:
            self.model.convert_to_fp16()

    def init_flow_warp(self, args):
        """Init the optical flow"""
        import argparse
        from raft import RAFT
        from .flow import load_img, get_flow

        video_frames_dir = self._output.video_frames_dir
        flo_fwd_dir = self._output.flo_fwd_dir
        flo_dir = self._output.flo_dir
        raft_model_path = self.get_model_path("RAFT")

        rsft_args = argparse.Namespace()
        rsft_args.small = False
        rsft_args.mixed_precision = True

        if not args.video_init_flow_warp:
            log.info("video_init_flow_warp not set, skipping")

        else:
            flows = glob(flo_dir + "/*.*")
            if (len(flows) > 0) and not args.force_flow_generation:
                log.info(
                    f"Skipping flow generation:\nFound {len(flows)} existing flow files in current working folder: {self._output.flo_dir}.\nIf you wish to generate new flow files, set force_flow_generation=True and run again."
                )

            if (len(flows) == 0) or args.force_flow_generation:
                frames = sorted(glob(video_frames_dir + "/*.*"))
                if len(frames) < 2:
                    log.warning(
                        f"WARNING!\nCannot create flow maps: Found {len(frames)} frames extracted from your video input.\nPlease check your video path."
                    )
                if len(frames) >= 2:

                    raft_model = torch.nn.DataParallel(RAFT(rsft_args))
                    raft_model.load_state_dict(torch.load(raft_model_path))
                    raft_model = raft_model.module.cuda(device=self.cuda_device).eval()

                    for f in Path(f"{flo_fwd_dir}").glob("*.*"):
                        f.unlink()

                    # TBD Call out to a consistency checker?
                    self.framecount = 0
                    for frame1, frame2 in tqdm(
                        zip(frames[:-1], frames[1:]), total=len(frames) - 1
                    ):

                        out_flow21_fn = f"{flo_fwd_dir}/{frame1.split('/')[-1]}"

                        frame1 = load_img(frame1, args.width_height)
                        frame2 = load_img(frame2, args.width_height)

                        flow21 = get_flow(frame2, frame1, raft_model)
                        np.save(out_flow21_fn, flow21)

                        if args.video_init_check_consistency:
                            # TBD
                            pass

                    del raft_model
                    gc.collect()

    def load_models(self):
        log.info("> loading diffusion models...")
        self.load_diffusion_models()
        log.info("> loading clip models...")
        self.load_clip_models()

    def load_diffusion_models(self):
        from guided_diffusion.script_util import model_and_diffusion_defaults
        from .secondary import SecondaryDiffusionImageNet2

        DEVICE = torch.device(
            f"cuda:{self.args.cuda_device}" if torch.cuda.is_available() else "cpu"
        )
        log.info(f"Using device:{DEVICE}")
        self.cuda_device = DEVICE

        if torch.cuda.is_available():
            # A100 fix thanks to Emad
            if torch.cuda.get_device_capability(self.cuda_device) == (8, 0):
                log.info("Disabling CUDNN for A100 gpu")
                torch.backends.cudnn.enabled = False

        self._model_and_diffusion_args = model_and_diffusion_defaults()
        self.diffusion_model = self.config.models.diffusion_model
        if self.diffusion_model in self.diffusion_model_config:
            cfg = self.diffusion_model_config[self.diffusion_model]
        else:
            cfg = self.diffusion_model_config["custom"]
        log.info(f"Loading diffusion model: {self.diffusion_model}")
        self._model_and_diffusion_args.update(eKonf.to_dict(cfg))
        self.model_default = self._model_and_diffusion_args["image_size"]

        if self.config.use_secondary_model:
            secondary_model_path = self.get_model_path("model_secondary")
            log.info(f"Loading secondary model: {secondary_model_path}")
            self.secondary_model = SecondaryDiffusionImageNet2()
            self.secondary_model.load_state_dict(
                torch.load(secondary_model_path, map_location="cpu")
            )
            self.secondary_model.eval().requires_grad_(False).to(self.cuda_device)
        else:
            self.secondary_model = None

    def load_clip_models(self):
        """Load the clips"""
        import torchvision.transforms as T
        import lpips
        import clip
        import open_clip

        _clips = self.config.models.clip_models
        clip_models = []
        for name, _use in _clips.items():
            if _use:
                model_name = self.model_map.clip_models[name]
                clip_models.append(
                    clip.load(model_name, jit=False)[0]
                    .eval()
                    .requires_grad_(False)
                    .to(self.cuda_device)
                )
        _clips = self.config.models.openclip_models
        for name, _use in _clips.items():
            if _use:
                model_args = self.model_map.clip_models[name]
                clip_models.append(
                    open_clip.create_model(**model_args)
                    .eval()
                    .requires_grad_(False)
                    .to(self.cuda_device)
                )
        self.clip_models = clip_models
        self.normalize = T.Normalize(
            mean=[0.48145466, 0.4578275, 0.40821073],
            std=[0.26862954, 0.26130258, 0.27577711],
        )
        self.lpips_model = lpips.LPIPS(net="vgg").to(self.cuda_device)

    def download_models(self):
        """Download the models"""
        download = self.args.download
        check_model_SHA = download.check_model_SHA
        for name, model in download.models.items():
            if not isinstance(model, str):
                downloaded = model.get("downloaded", True)
                if downloaded:
                    log.info(f"Downloading model {name} from {model}")
                    _download_models(name, **model, check_model_SHA=check_model_SHA)
                else:
                    log.info(f"Skipping download of {name}")

        if not os.path.exists(download.pretrained_symlink):
            os.symlink(download.pretrained_dir, download.pretrained_symlink)
            log.info(
                f"Created symlink of {download.pretrained_dir} to {download.pretrained_symlink}"
            )

    def _3d_step(self, img_filepath, frame_num, midas_model, midas_transform, args):
        import py3d_tools as p3dT
        from . import disco_xform_utils as dxf

        if args.key_frames:
            translation_x = self.translation_x_series[frame_num]
            translation_y = self.translation_y_series[frame_num]
            translation_z = self.translation_z_series[frame_num]
            rotation_3d_x = self.rotation_3d_x_series[frame_num]
            rotation_3d_y = self.rotation_3d_y_series[frame_num]
            rotation_3d_z = self.rotation_3d_z_series[frame_num]
            log.info(f"translation_x: {translation_x}")
            log.info(f"translation_y: {translation_y}")
            log.info(f"translation_z: {translation_z}")
            log.info(f"rotation_3d_x: {rotation_3d_x}")
            log.info(f"rotation_3d_y: {rotation_3d_y}")
            log.info(f"rotation_3d_z: {rotation_3d_z}")

        translate_xyz = [
            -translation_x * self.TRANSLATION_SCALE,
            translation_y * self.TRANSLATION_SCALE,
            -translation_z * self.TRANSLATION_SCALE,
        ]
        rotate_xyz_degrees = [rotation_3d_x, rotation_3d_y, rotation_3d_z]
        log.info(f"translation: {translate_xyz}")
        log.info(f"rotation: {rotate_xyz_degrees}")
        rotate_xyz = [
            math.radians(rotate_xyz_degrees[0]),
            math.radians(rotate_xyz_degrees[1]),
            math.radians(rotate_xyz_degrees[2]),
        ]
        rot_mat = p3dT.euler_angles_to_matrix(
            torch.tensor(rotate_xyz, device=self.cuda_device), "XYZ"
        ).unsqueeze(0)
        log.info(f"rot_mat: {str(rot_mat)}")
        next_step_pil = dxf.transform_image_3d(
            img_filepath,
            midas_model,
            midas_transform,
            self.cuda_device,
            rot_mat,
            translate_xyz,
            args.near_plane,
            args.far_plane,
            args.fov,
            padding_mode=args.padding_mode,
            sampling_mode=args.sampling_mode,
            midas_weight=args.midas_weight,
        )
        return next_step_pil

    def _get_model_stats(self, text_prompt, image_prompt, args):
        """Update the model stats"""
        import clip
        from .diffuse import parse_prompt, MakeCutouts, fetch

        model_stats = []
        weights = []

        for clip_model in self.clip_models:
            cutn = 16
            model_stat = {
                "clip_model": None,
                "target_embeds": [],
                "make_cutouts": None,
                "weights": [],
            }
            model_stat["clip_model"] = clip_model

            for prompt in text_prompt:
                txt, weight = parse_prompt(prompt)
                txt = clip_model.encode_text(
                    clip.tokenize(prompt).to(self.cuda_device)
                ).float()

                if args.fuzzy_prompt:
                    for i in range(25):
                        model_stat["target_embeds"].append(
                            (txt + torch.randn(txt.shape).cuda() * args.rand_mag).clamp(
                                0, 1
                            )
                        )
                        model_stat["weights"].append(weight)
                else:
                    model_stat["target_embeds"].append(txt)
                    model_stat["weights"].append(weight)

            if image_prompt:
                model_stat["make_cutouts"] = MakeCutouts(
                    clip_model.visual.input_resolution,
                    cutn,
                    skip_augs=args.skip_augs,
                )
                for prompt in image_prompt:
                    path, weight = parse_prompt(prompt)
                    img = Image.open(fetch(path)).convert("RGB")
                    img = TF.resize(
                        img,
                        min(args.side_x, args.side_y, *img.size),
                        T.InterpolationMode.LANCZOS,
                    )
                    batch = model_stat["make_cutouts"](
                        TF.to_tensor(img)
                        .to(self.cuda_device)
                        .unsqueeze(0)
                        .mul(2)
                        .sub(1)
                    )
                    embed = clip_model.encode_image(self.normalize(batch)).float()
                    if args.fuzzy_prompt:
                        for i in range(25):
                            model_stat["target_embeds"].append(
                                (
                                    embed
                                    + torch.randn(embed.shape).cuda() * args.rand_mag
                                ).clamp(0, 1)
                            )
                            weights.extend([weight / cutn] * cutn)
                    else:
                        model_stat["target_embeds"].append(embed)
                        model_stat["weights"].extend([weight / cutn] * cutn)

            model_stat["target_embeds"] = torch.cat(model_stat["target_embeds"])
            model_stat["weights"] = torch.tensor(
                model_stat["weights"], device=self.cuda_device
            )
            if model_stat["weights"].sum().abs() < 1e-3:
                raise RuntimeError("The weights must not sum to 0.")
            model_stat["weights"] /= model_stat["weights"].sum().abs()
            model_stats.append(model_stat)

        return model_stats

    def _init_image(self, init_image, args):
        from .diffuse import create_perlin_noise, fetch

        init = None
        if init_image is not None:
            init = Image.open(fetch(init_image)).convert("RGB")
            init = init.resize((args.side_x, args.side_y), Image.LANCZOS)
            init = TF.to_tensor(init).to(self.cuda_device).unsqueeze(0).mul(2).sub(1)

        if args.perlin_init:
            log.info(f"Perlin init with {args.perlin_mode} mode")
            if args.perlin_mode == "color":
                grayscale = False
                grayscale2 = False
            elif args.perlin_mode == "gray":
                grayscale = True
                grayscale2 = True
            else:
                grayscale = False
                grayscale2 = True
            init = create_perlin_noise(
                args.side_x,
                args.side_y,
                octaves=[1.5**-i * 0.5 for i in range(12)],
                width=1,
                height=1,
                grayscale=grayscale,
                device=self.cuda_device,
            )
            init2 = create_perlin_noise(
                args.side_x,
                args.side_y,
                octaves=[1.5**-i * 0.5 for i in range(8)],
                width=4,
                height=4,
                grayscale=grayscale2,
                device=self.cuda_device,
            )
            # init = TF.to_tensor(init).add(TF.to_tensor(init2)).div(2).to(device)
            init = (
                TF.to_tensor(init)
                .add(TF.to_tensor(init2))
                .div(2)
                .to(self.cuda_device)
                .unsqueeze(0)
                .mul(2)
                .sub(1)
            )
            del init2
        return init

    def _progress_samples(
        self,
        cur_t,
        init,
        init_scale,
        skip_steps,
        loss_values,
        model_stats,
        frame_num,
        diffusion,
        midas_model,
        midas_transform,
        args,
    ):
        import cv2
        from .secondary import alpha_sigma_to_t
        from .diffuse import (
            MakeCutoutsDango,
            spherical_dist_loss,
            tv_loss,
            range_loss,
            regen_perlin,
            generate_eye_views,
        )

        batch_dir = self._output.batch_dir
        partial_dir = self._output.partial_dir
        prev_frame_path = self._output.prev_frame_path
        prev_frame_scaled_path = self._output.prev_frame_scaled_path
        progress_path = self._output.progress_path

        secondary_model = self.secondary_model
        model = self.model
        if isinstance(args.cut_overview, str):
            cut_overview = eval(args.cut_overview)
            cut_innercut = eval(args.cut_innercut)
            cut_ic_pow = eval(args.cut_ic_pow)
            cut_icgray_p = eval(args.cut_icgray_p)
        else:
            cut_overview = args.cut_overview
            cut_innercut = args.cut_innercut
            cut_ic_pow = args.cut_ic_pow
            cut_icgray_p = args.cut_icgray_p

        def cond_fn(x, t, y=None):
            with torch.enable_grad():
                x_is_NaN = False
                x = x.detach().requires_grad_()
                n = x.shape[0]
                if args.use_secondary_model:
                    alpha = torch.tensor(
                        diffusion.sqrt_alphas_cumprod[cur_t],
                        device=self.cuda_device,
                        dtype=torch.float32,
                    )
                    sigma = torch.tensor(
                        diffusion.sqrt_one_minus_alphas_cumprod[cur_t],
                        device=self.cuda_device,
                        dtype=torch.float32,
                    )
                    cosine_t = alpha_sigma_to_t(alpha, sigma)
                    out = secondary_model(x, cosine_t[None].repeat([n])).pred
                    fac = diffusion.sqrt_one_minus_alphas_cumprod[cur_t]
                    x_in = out * fac + x * (1 - fac)
                    x_in_grad = torch.zeros_like(x_in)
                else:
                    my_t = (
                        torch.ones([n], device=self.cuda_device, dtype=torch.long)
                        * cur_t
                    )
                    out = diffusion.p_mean_variance(
                        model, x, my_t, clip_denoised=False, model_kwargs={"y": y}
                    )
                    fac = diffusion.sqrt_one_minus_alphas_cumprod[cur_t]
                    x_in = out["pred_xstart"] * fac + x * (1 - fac)
                    x_in_grad = torch.zeros_like(x_in)

                for model_stat in model_stats:
                    for i in range(args.cutn_batches):
                        t_int = (
                            int(t.item()) + 1
                        )  # errors on last step without +1, need to find source
                        # when using SLIP Base model the dimensions need to be hard coded to avoid AttributeError: 'VisionTransformer' object has no attribute 'input_resolution'
                        try:
                            input_resolution = model_stat[
                                "clip_model"
                            ].visual.input_resolution
                        except:
                            input_resolution = 224

                        cuts = MakeCutoutsDango(
                            args.animation_mode,
                            input_resolution,
                            args.skip_augs,
                            Overview=cut_overview[1000 - t_int],
                            InnerCrop=cut_innercut[1000 - t_int],
                            IC_Size_Pow=cut_ic_pow[1000 - t_int],
                            IC_Grey_P=cut_icgray_p[1000 - t_int],
                        )
                        clip_in = self.normalize(cuts(x_in.add(1).div(2)))
                        image_embeds = (
                            model_stat["clip_model"].encode_image(clip_in).float()
                        )
                        dists = spherical_dist_loss(
                            image_embeds.unsqueeze(1),
                            model_stat["target_embeds"].unsqueeze(0),
                        )
                        dists = dists.view(
                            [
                                cut_overview[1000 - t_int] + cut_innercut[1000 - t_int],
                                n,
                                -1,
                            ]
                        )
                        losses = dists.mul(model_stat["weights"]).sum(2).mean(0)
                        loss_values.append(
                            losses.sum().item()
                        )  # log loss, probably shouldn't do per cutn_batch
                        x_in_grad += (
                            torch.autograd.grad(
                                losses.sum() * args.clip_guidance_scale, x_in
                            )[0]
                            / args.cutn_batches
                        )
                tv_losses = tv_loss(x_in)
                if args.use_secondary_model:
                    range_losses = range_loss(out)
                else:
                    range_losses = range_loss(out["pred_xstart"])
                sat_losses = torch.abs(x_in - x_in.clamp(min=-1, max=1)).mean()
                loss = (
                    tv_losses.sum() * args.tv_scale
                    + range_losses.sum() * args.range_scale
                    + sat_losses.sum() * args.sat_scale
                )
                if init is not None and init_scale:
                    init_losses = self.lpips_model(x_in, init)
                    loss = loss + init_losses.sum() * init_scale
                x_in_grad += torch.autograd.grad(loss, x_in)[0]
                if torch.isnan(x_in_grad).any() == False:
                    grad = -torch.autograd.grad(x_in, x, x_in_grad)[0]
                else:
                    # print("NaN'd")
                    x_is_NaN = True
                    grad = torch.zeros_like(x)
            if args.clamp_grad and x_is_NaN == False:
                magnitude = grad.square().mean().sqrt()
                return (
                    grad * magnitude.clamp(max=args.clamp_max) / magnitude
                )  # min=-0.02, min=-clamp_max,
            return grad

        def symmetry_transformation_fn(x):
            if args.use_horizontal_symmetry:
                [n, c, h, w] = x.size()
                x = torch.concat(
                    (x[:, :, :, : w // 2], torch.flip(x[:, :, :, : w // 2], [-1])),
                    -1,
                )
                log.info("horizontal symmetry applied")
            if args.use_vertical_symmetry:
                [n, c, h, w] = x.size()
                x = torch.concat(
                    (x[:, :, : h // 2, :], torch.flip(x[:, :, : h // 2, :], [-2])),
                    -2,
                )
                log.info("vertical symmetry applied")
            return x

        if args.diffusion_sampling_mode == "ddim":
            sampling_fn = diffusion.ddim_sample_loop_progressive
        else:
            sampling_fn = diffusion.plms_sample_loop_progressive

        if args.perlin_init:
            init = regen_perlin(
                args.perlin_mode,
                args.batch_size,
                args.side_x,
                args.side_y,
                self.cuda_device,
            )

        if args.diffusion_sampling_mode == "ddim":
            progress_samples = sampling_fn(
                model,
                (args.batch_size, 3, args.side_y, args.side_x),
                clip_denoised=args.clip_denoised,
                model_kwargs={},
                cond_fn=cond_fn,
                progress=True,
                skip_timesteps=skip_steps,
                init_image=init,
                randomize_class=args.randomize_class,
                eta=args.eta,
                transformation_fn=symmetry_transformation_fn,
                transformation_percent=args.transformation_percent,
            )
        else:
            progress_samples = sampling_fn(
                model,
                (args.batch_size, 3, args.side_y, args.side_x),
                clip_denoised=args.clip_denoised,
                model_kwargs={},
                cond_fn=cond_fn,
                progress=True,
                skip_timesteps=skip_steps,
                init_image=init,
                randomize_class=args.randomize_class,
                order=2,
            )

        for step_num, samples in enumerate(progress_samples):
            cur_t -= 1
            intermediate_step = False
            if args.steps_per_checkpoint is not None:
                if step_num % args.steps_per_checkpoint == 0 and step_num > 0:
                    intermediate_step = True
            elif step_num in args.intermediate_saves:
                intermediate_step = True
            res = {
                "sample_num": frame_num,
                "step": step_num,
                "cur_t": cur_t,
                "image": None,
                "pregress_path": None,
            }

            if step_num % args.display_rate == 0 or cur_t == -1 or intermediate_step:
                for k, image in enumerate(samples["pred_xstart"]):
                    # tqdm.write(f'Batch {i}, step {j}, output {k}:')
                    res["ouput"] = k
                    if args.n_samples > 0:
                        # if intermediates are saved to the subfolder, don't append a step or percentage to the name
                        if cur_t == -1 and args.intermediates_in_subfolder:
                            filename = f"{args.batch_name}({args.batch_num})_{frame_num:04}.png"
                        else:
                            filename = f"{args.batch_name}({args.batch_num})_{frame_num:04}-{step_num:03}.png"
                    image = TF.to_pil_image(image.add(1).div(2).clamp(0, 1))
                    if step_num % args.display_rate == 0 or cur_t == -1:
                        image.save(progress_path)
                        res["image"] = image
                        res["pregress_path"] = progress_path

                    if args.intermediates_in_subfolder:
                        _img_path = os.path.join(partial_dir, filename)
                    else:
                        _img_path = os.path.join(batch_dir, filename)
                    res["image_path"] = _img_path

                    if args.steps_per_checkpoint is not None:
                        if step_num % args.steps_per_checkpoint == 0 and step_num > 0:
                            image.save(_img_path)
                    else:
                        if step_num in args.intermediate_saves:
                            image.save(_img_path)
                    if cur_t == -1:
                        if args.animation_mode != AnimMode.NONE:
                            image.save(prev_frame_path)
                        _img_path = os.path.join(batch_dir, filename)
                        image.save(_img_path)
                        self.sample_imagepaths.append(_img_path)
                        log.info(f"Saved {filename}")
                        if args.animation_mode == AnimMode.ANIM_3D:
                            # If turbo, save a blended image
                            _img_path = os.path.join(batch_dir, filename)
                            if args.turbo_mode and frame_num > 0:
                                # Mix new image with prevFrameScaled
                                blend_factor = (1) / int(args.turbo_steps)
                                # This is already updated..
                                new_frame = cv2.imread(prev_frame_path)
                                prev_frame_warped = cv2.imread(prev_frame_scaled_path)
                                blended_image = cv2.addWeighted(
                                    new_frame,
                                    blend_factor,
                                    prev_frame_warped,
                                    (1 - blend_factor),
                                    0.0,
                                )
                                cv2.imwrite(_img_path, blended_image)
                            else:
                                image.save(_img_path)

                            if args.vr_mode:
                                generate_eye_views(
                                    self.TRANSLATION_SCALE,
                                    batch_dir,
                                    filename,
                                    frame_num,
                                    midas_model,
                                    midas_transform,
                                    self.cuda_device,
                                    args.vr_eye_angle,
                                    args.vr_ipd,
                                    args.near_plane,
                                    args.far_plane,
                                    args.fov,
                                    args.padding_mode,
                                    args.sampling_mode,
                                    args.midas_weight,
                                )
                    yield res
            else:
                yield res

    def _generate_progress_samples(
        self,
        cur_t,
        init,
        init_scale,
        skip_steps,
        loss_values,
        model_stats,
        frame_num,
        diffusion,
        midas_model,
        midas_transform,
        image_display,
        args,
    ):
        progress_samples = self._progress_samples(
            cur_t,
            init,
            init_scale,
            skip_steps,
            loss_values,
            model_stats,
            frame_num,
            diffusion,
            midas_model,
            midas_transform,
            args,
        )
        for sample in progress_samples:
            progress_path = sample.get("pregress_path")
            if image_display is not None:
                with image_display:
                    if progress_path is not None:
                        eKonf.clear_output(wait=True)
                        eKonf.display_image(progress_path)
                # print(sample)

        return loss_values

    def _init_2d(
        self,
        frame_num,
        seed,
        args,
    ):
        import cv2

        batch_dir = self._output.batch_dir
        prev_frame_path = self._output.prev_frame_path
        prev_frame_scaled_path = self._output.prev_frame_scaled_path

        if args.key_frames:
            angle = self.angle_series[frame_num]
            zoom = self.zoom_series[frame_num]
            translation_x = self.translation_x_series[frame_num]
            translation_y = self.translation_y_series[frame_num]
            log.info(
                f"angle: {angle}, zoom: {zoom}, translation_x: {translation_x}, translation_y: {translation_y}"
            )

        seed += 1
        if args.resume_run and frame_num == args.start_frame:
            _img_path = os.path.join(
                batch_dir,
                f"{args.batch_name}({args.batch_num})_{args.start_frame-1:04}.png",
            )
            img_0 = cv2.imread(_img_path)
        else:
            img_0 = cv2.imread(prev_frame_path)
        center = (1 * img_0.shape[1] // 2, 1 * img_0.shape[0] // 2)
        trans_mat = np.float32([[1, 0, translation_x], [0, 1, translation_y]])
        rot_mat = cv2.getRotationMatrix2D(center, angle, zoom)
        trans_mat = np.vstack([trans_mat, [0, 0, 1]])
        rot_mat = np.vstack([rot_mat, [0, 0, 1]])
        transformation_matrix = np.matmul(rot_mat, trans_mat)
        img_0 = cv2.warpPerspective(
            img_0,
            transformation_matrix,
            (img_0.shape[1], img_0.shape[0]),
            borderMode=cv2.BORDER_WRAP,
        )

        init_image = prev_frame_scaled_path
        cv2.imwrite(init_image, img_0)
        init_scale = args.frames_scale
        skip_steps = args.calc_frames_skip_steps
        return seed, init_image, init_scale, skip_steps

    def _init_3d(
        self,
        frame_num,
        seed,
        midas_model,
        midas_transform,
        args,
    ):
        import cv2
        from .diffuse import generate_eye_views

        batch_dir = self._output.batch_dir
        prev_frame_path = self._output.prev_frame_path
        prev_frame_scaled_path = self._output.prev_frame_scaled_path
        old_frame_scaled_path = self._output.old_frame_scaled_path
        skip_frame = False

        seed += 1
        if args.resume_run and frame_num == args.start_frame:
            img_filepath = os.path.join(
                batch_dir,
                f"{args.batch_name}({args.batch_num})_{args.start_frame-1:04}.png",
            )
            if args.turbo_mode and frame_num > args.turbo_preroll:
                shutil.copyfile(img_filepath, old_frame_scaled_path)
        else:
            img_filepath = prev_frame_path

        next_step_pil = self._3d_step(
            img_filepath, frame_num, midas_model, midas_transform, args
        )
        next_step_pil.save(prev_frame_scaled_path)

        # Turbo mode - skip some diffusions, use 3d morph for clarity and to save time
        if args.turbo_mode:
            if frame_num == args.turbo_preroll:  # start tracking oldframe
                # stash for later blending
                next_step_pil.save(old_frame_scaled_path)
            elif frame_num > args.turbo_preroll:
                # set up 2 warped image sequences, old & new, to blend toward new diff image
                old_frame = self._3d_step(
                    old_frame_scaled_path,
                    frame_num,
                    midas_model,
                    midas_transform,
                    args,
                )
                old_frame.save(old_frame_scaled_path)
                if frame_num % int(args.turbo_steps) != 0:
                    log.info("turbo skip this frame: skipping clip diffusion steps")
                    filename = f"{args.batch_name}({args.batch_num})_{frame_num:04}.png"
                    blend_factor = ((frame_num % int(args.turbo_steps)) + 1) / int(
                        args.turbo_steps
                    )
                    log.info(
                        "turbo skip this frame: skipping clip diffusion steps and saving blended frame"
                    )
                    # this is already updated..
                    newWarpedImg = cv2.imread(prev_frame_scaled_path)
                    oldWarpedImg = cv2.imread(old_frame_scaled_path)
                    blendedImage = cv2.addWeighted(
                        newWarpedImg,
                        blend_factor,
                        oldWarpedImg,
                        1 - blend_factor,
                        0.0,
                    )
                    _img_path = os.path.join(batch_dir, filename)
                    cv2.imwrite(_img_path, blendedImage)
                    # save it also as prev_frame to feed next iteration
                    next_step_pil.save(img_filepath)
                    if args.vr_mode:
                        generate_eye_views(
                            self.TRANSLATION_SCALE,
                            self._output.batch_dir,
                            filename,
                            frame_num,
                            midas_model,
                            midas_transform,
                            self.cuda_device,
                            args.vr_eye_angle,
                            args.vr_ipd,
                            args.near_plane,
                            args.far_plane,
                            args.fov,
                            args.padding_mode,
                            args.sampling_mode,
                            args.midas_weight,
                        )
                    skip_frame = True
                else:
                    # if not a skip frame, will run diffusion and need to blend.
                    oldWarpedImg = cv2.imread(prev_frame_scaled_path)
                    # swap in for blending later
                    cv2.imwrite(old_frame_scaled_path, oldWarpedImg)
                    log.info("clip/diff this frame - generate clip diff image")

        init_image = prev_frame_scaled_path
        init_scale = args.frames_scale
        skip_steps = args.calc_frames_skip_steps

        return seed, init_image, init_scale, skip_steps, skip_frame

    def _init_video(self, frame_num, seed, args):
        from .flow import warp

        batch_dir = self._output.batch_dir
        video_frames_dir = self._output.video_frames_dir
        warped_path = self._output.warped_path
        flo_dir = self._output.flo_dir

        init_scale = args.video_init_frames_scale
        skip_steps = args.calc_frames_skip_steps

        if not args.video_init_seed_continuity:
            seed += 1
        if args.video_init_flow_warp:
            if frame_num == 0:
                skip_steps = args.video_init_skip_steps
                init_image = os.path.join(video_frames_dir, f"{frame_num+1:04}.jpg")
            if frame_num > 0:
                _img_path = os.path.join(
                    batch_dir,
                    f"{args.batch_name}({args.batch_num})_{frame_num-1:04}.png",
                )
                prev = PIL.Image.open(_img_path)

                frame1_path = os.path.join(video_frames_dir, f"{frame_num:04}.jpg")
                frame2_path = os.path.join(video_frames_dir, f"{frame_num+1:04}.jpg")
                frame2 = PIL.Image.open(frame2_path)
                flo_path = os.path.join(
                    flo_dir,
                    f"{frame1_path.split('/')[-1]}.npy",
                )

                init_image = warped_path
                log.info(
                    f"warping frames with flow blend ratio: {args.video_init_flow_blend}"
                )
                weights_path = None
                if args.video_init_check_consistency:
                    # TBD
                    pass

                warp(
                    prev,
                    frame2,
                    flo_path,
                    blend=args.video_init_flow_blend,
                    weights_path=weights_path,
                ).save(init_image)

        else:
            init_image = os.path.join(video_frames_dir, f"{frame_num+1:04}.jpg")
        return seed, init_image, init_scale, skip_steps

    def _run_anim(self, args):
        """Run the simulation"""
        from .midas import Midas

        seed = args.seed
        diffusion = self.diffusion
        midas_model = None
        midas_transform = None
        cur_t = None
        batchBar = None

        text_series, image_series = self._get_prompt_series(args, args.max_frames)

        if (args.animation_mode == AnimMode.ANIM_3D) and (args.midas_weight > 0.0):
            midas = Midas(**self.midas_config)
            midas_model, midas_transform, _, _, _, _ = midas.init_midas_depth_model(
                args.midas_depth_model
            )

        log.info(f"looping over range({args.start_frame}, {args.max_frames})")
        for frame_num in range(args.start_frame, args.max_frames):
            # Make sure GPU memory doesn't get corrupted from cancelling the run mid-way through, allow a full frame to complete
            if args.stop_on_next_loop:
                break

            eKonf.clear_output(wait=True)
            batchBar = tqdm(
                range(args.max_frames),
                desc=f"{args.batch_name}({args.batch_num}) frames",
            )
            batchBar.n = frame_num
            batchBar.refresh()

            init_image = None
            if eKonf.exists(args.init_image):
                init_image = args.init_image
            init_scale = args.init_scale
            skip_steps = args.skip_steps

            if args.animation_mode == AnimMode.ANIM_2D:
                if frame_num > 0:
                    seed, init_image, init_scale, skip_steps = self._init_2d(
                        frame_num,
                        seed,
                        args,
                    )
            if args.animation_mode == AnimMode.ANIM_3D:
                if frame_num > 0:
                    (
                        seed,
                        init_image,
                        init_scale,
                        skip_steps,
                        skip_frame,
                    ) = self._init_3d(
                        frame_num,
                        seed,
                        midas_model,
                        midas_transform,
                        args,
                    )
                    if skip_frame:
                        continue
            if args.animation_mode == AnimMode.VIDEO_INPUT:
                seed, init_image, init_scale, skip_steps = self._init_video(
                    frame_num, seed, args
                )
            log.info(f"init_image: {init_image}")
            log.info(
                f"init_scale: {init_scale}, skip_steps: {skip_steps}, seed: {seed}"
            )

            loss_values = []

            if seed is not None:
                np.random.seed(seed)
                random.seed(seed)
                torch.manual_seed(seed)
                torch.cuda.manual_seed_all(seed)
                torch.backends.cudnn.deterministic = True

            text_prompt, image_prompt = self._get_prompt(
                text_series, image_series, frame_num
            )
            log.info(f"Image prompt: {image_prompt}")
            log.info(f"Frame {frame_num} Prompt: {text_prompt}")

            model_stats = self._get_model_stats(text_prompt, image_prompt, args)
            init = self._init_image(init_image, args)

            image_display = eKonf.get_display()

            eKonf.display(image_display)
            gc.collect()
            torch.cuda.empty_cache()
            cur_t = diffusion.num_timesteps - skip_steps - 1

            self._generate_progress_samples(
                cur_t,
                init,
                init_scale,
                skip_steps,
                loss_values,
                model_stats,
                frame_num,
                diffusion,
                midas_model,
                midas_transform,
                image_display,
                args,
            )

            if args.get("show_losses", False):
                plt.plot(np.array(loss_values), "r")

        if batchBar is not None:
            batchBar.n = frame_num + 1
            batchBar.refresh()

    def _get_prompt_series(self, args, max_frames):
        text_prompts = eKonf.to_dict(args.text_prompts)
        image_prompts = eKonf.to_dict(args.image_prompts)
        text_series = split_prompts(text_prompts, max_frames) if text_prompts else None
        image_series = (
            split_prompts(image_prompts, max_frames) if image_prompts else None
        )
        return text_series, image_series

    def _get_prompt(self, text_series, image_series, frame_num):
        if text_series is not None and frame_num >= len(text_series):
            text_prompt = text_series[-1]
        elif text_series is not None:
            text_prompt = text_series[frame_num]
        else:
            text_prompt = []

        if image_series is not None and frame_num >= len(image_series):
            image_prompt = image_series[-1]
        elif image_series is not None:
            image_prompt = image_series[frame_num]
        else:
            image_prompt = []
        return text_prompt, image_prompt

    def _run(self, args):
        """Run the simulation"""
        seed = args.seed
        diffusion = self.diffusion

        midas_model = None
        midas_transform = None
        init_image = None
        cur_t = None
        batchBar = None
        loss_values = []

        text_series, image_series = self._get_prompt_series(args, args.n_samples)

        if eKonf.exists(args.init_image):
            init_image = args.init_image
        init_scale = args.init_scale
        skip_steps = args.skip_steps
        log.info(f"init_image: {init_image}")
        log.info(f"init_scale: {init_scale}, skip_steps: {skip_steps}, seed: {seed}")

        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.deterministic = True

        image_display = eKonf.get_display()
        log.info(f"looping over range({args.start_sample}, {args.n_samples})")
        for sample_num in range(args.start_sample, args.n_samples):
            # Make sure GPU memory doesn't get corrupted from cancelling the run mid-way through, allow a full frame to complete
            if args.stop_on_next_loop:
                break

            eKonf.clear_output(wait=True)
            batchBar = tqdm(
                range(args.n_samples),
                desc=f"{args.batch_name}({args.batch_num}) samples",
            )
            batchBar.n = sample_num
            batchBar.refresh()

            text_prompt, image_prompt = self._get_prompt(
                text_series, image_series, sample_num
            )
            log.info(f"Image prompt: {image_prompt}")
            log.info(f"Sample {sample_num} Prompt: {text_prompt}")

            model_stats = self._get_model_stats(text_prompt, image_prompt, args)
            init = self._init_image(init_image, args)

            eKonf.display(image_display)
            gc.collect()
            torch.cuda.empty_cache()
            cur_t = diffusion.num_timesteps - skip_steps - 1

            self._generate_progress_samples(
                cur_t,
                init,
                init_scale,
                skip_steps,
                loss_values,
                model_stats,
                sample_num,
                diffusion,
                midas_model,
                midas_transform,
                image_display,
                args,
            )

        if batchBar is not None:
            batchBar.n = sample_num + 1
            batchBar.refresh()
            if args.get("show_losses", False):
                plt.plot(np.array(loss_values), "r")
        return loss_values

    def _generator(self, args):
        """Run the simulation"""
        seed = args.seed
        diffusion = self.diffusion

        midas_model = None
        midas_transform = None
        init_image = None
        cur_t = None
        loss_values = []

        text_series, image_series = self._get_prompt_series(args, args.n_samples)

        if eKonf.exists(args.init_image):
            init_image = args.init_image
        init_scale = args.init_scale
        skip_steps = args.skip_steps
        log.info(f"init_image: {init_image}")
        log.info(f"init_scale: {init_scale}, skip_steps: {skip_steps}, seed: {seed}")

        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.deterministic = True

        for sample_num in range(args.start_sample, args.n_samples):
            # Make sure GPU memory doesn't get corrupted from cancelling the run mid-way through, allow a full frame to complete
            if args.stop_on_next_loop:
                break

            text_prompt, image_prompt = self._get_prompt(
                text_series, image_series, sample_num
            )
            log.info(f"Image prompt: {image_prompt}")
            log.info(f"Sample {sample_num} Prompt: {text_prompt}")

            model_stats = self._get_model_stats(text_prompt, image_prompt, args)
            init = self._init_image(init_image, args)

            gc.collect()
            torch.cuda.empty_cache()
            cur_t = diffusion.num_timesteps - skip_steps - 1

            for sample in self._progress_samples(
                cur_t,
                init,
                init_scale,
                skip_steps,
                loss_values,
                model_stats,
                sample_num,
                diffusion,
                midas_model,
                midas_transform,
                args,
            ):
                yield sample

    def _prepare_config(self, args):
        batch_dir = self._output.batch_dir
        video_frames_dir = self._output.video_frames_dir
        flo_dir = self._output.flo_dir

        # Ensure prompts formatted correctly
        text_prompts = eKonf.to_dict(args.text_prompts)
        image_prompts = eKonf.to_dict(args.image_prompts)
        if isinstance(text_prompts, str):
            text_prompts = {0: [text_prompts]}
            log.info(f"converted string type text_prompts to {text_prompts}")
        elif isinstance(text_prompts, list):
            text_prompts = {0: text_prompts}
            log.info(f"converted list type text_prompts to {text_prompts}")
        if isinstance(image_prompts, str):
            image_prompts = {0: [image_prompts]}
            log.info(f"converted string type image_prompts to {image_prompts}")
        elif isinstance(image_prompts, list):
            image_prompts = {0: image_prompts}
            log.info(f"converted list type image_prompts to {image_prompts}")
        args.text_prompts = text_prompts
        args.image_prompts = image_prompts

        args.width_height = (
            args.width_height_for_256x256_models
            if args.models.diffusion_model
            in self.model_map.diffusion_models_256x256_list
            else args.width_height_for_512x512_models
        )
        log.info(
            f"width_height: {args.width_height}, diffusion_model: {args.models.diffusion_model}"
        )

        # Get corrected sizes
        args.side_x = (args.width_height[0] // 64) * 64
        args.side_y = (args.width_height[1] // 64) * 64
        if args.side_x != args.width_height[0] or args.side_y != args.width_height[1]:
            log.info(
                f"Changing output size to {args.side_x}x{args.side_y}. Dimensions must by multiples of 64."
            )

        _mode = AnimMode.NONE.value
        for mode in AnimMode:
            if mode.value.lower() == args.animation_mode.lower():
                _mode = mode.value
                break
        args.animation_mode = _mode

        if args.animation_mode == AnimMode.VIDEO_INPUT:
            args.steps = args.video_init_steps
            extract_frames(
                args.video_init_path,
                args.extract_nth_frame,
                video_frames_dir,
            )
            args.max_frames = len(glob(f"{video_frames_dir}/*.jpg"))

        # insist turbo be used only w 3d anim.
        if args.turbo_mode and args.animation_mode != AnimMode.ANIM_3D:
            log.info("Turbo mode only available with 3D animations. Disabling Turbo.")
            args.turbo_mode = False

        # insist VR be used only w 3d anim.
        if args.vr_mode and args.animation_mode != AnimMode.ANIM_3D:
            log.info("VR mode only available with 3D animations. Disabling VR.")
            args.vr_mode = False

        if args.key_frames:
            try:
                self.angle_series = get_inbetweens(
                    parse_key_frames(args.angle), args.max_frames, args.interp_spline
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `angle` correctly for key frames.\n"
                    "Attempting to interpret `angle` as "
                    f'"0: ({args.angle})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.angle = f"0: ({args.angle})"
                self.angle_series = get_inbetweens(
                    parse_key_frames(args.angle), args.max_frames, args.interp_spline
                )

            try:
                self.zoom_series = get_inbetweens(
                    parse_key_frames(args.zoom), args.max_frames, args.interp_spline
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `zoom` correctly for key frames.\n"
                    "Attempting to interpret `zoom` as "
                    f'"0: ({args.zoom})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.zoom = f"0: ({args.zoom})"
                self.zoom_series = get_inbetweens(
                    parse_key_frames(args.zoom), args.max_frames, args.interp_spline
                )

            try:
                self.translation_x_series = get_inbetweens(
                    parse_key_frames(args.translation_x),
                    args.max_frames,
                    args.interp_spline,
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `translation_x` correctly for key frames.\n"
                    "Attempting to interpret `translation_x` as "
                    f'"0: ({args.translation_x})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.translation_x = f"0: ({args.translation_x})"
                self.translation_x_series = get_inbetweens(
                    parse_key_frames(args.translation_x),
                    args.max_frames,
                    args.interp_spline,
                )

            try:
                self.translation_y_series = get_inbetweens(
                    parse_key_frames(args.translation_y),
                    args.max_frames,
                    args.interp_spline,
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `translation_y` correctly for key frames.\n"
                    "Attempting to interpret `translation_y` as "
                    f'"0: ({args.translation_y})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.translation_y = f"0: ({args.translation_y})"
                self.translation_y_series = get_inbetweens(
                    parse_key_frames(args.translation_y),
                    args.max_frames,
                    args.interp_spline,
                )

            try:
                self.translation_z_series = get_inbetweens(
                    parse_key_frames(args.translation_z),
                    args.max_frames,
                    args.interp_spline,
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `translation_z` correctly for key frames.\n"
                    "Attempting to interpret `translation_z` as "
                    f'"0: ({args.translation_z})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.translation_z = f"0: ({args.translation_z})"
                self.translation_z_series = get_inbetweens(
                    parse_key_frames(args.translation_z),
                    args.max_frames,
                    args.interp_spline,
                )

            try:
                self.rotation_3d_x_series = get_inbetweens(
                    parse_key_frames(args.rotation_3d_x),
                    args.max_frames,
                    args.interp_spline,
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `rotation_3d_x` correctly for key frames.\n"
                    "Attempting to interpret `rotation_3d_x` as "
                    f'"0: ({args.rotation_3d_x})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.rotation_3d_x = f"0: ({args.rotation_3d_x})"
                self.rotation_3d_x_series = get_inbetweens(
                    parse_key_frames(args.rotation_3d_x),
                    args.max_frames,
                    args.interp_spline,
                )

            try:
                self.rotation_3d_y_series = get_inbetweens(
                    parse_key_frames(args.rotation_3d_y),
                    args.max_frames,
                    args.interp_spline,
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `rotation_3d_y` correctly for key frames.\n"
                    "Attempting to interpret `rotation_3d_y` as "
                    f'"0: ({args.rotation_3d_y})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.rotation_3d_y = f"0: ({args.rotation_3d_y})"
                self.rotation_3d_y_series = get_inbetweens(
                    parse_key_frames(args.rotation_3d_y),
                    args.max_frames,
                    args.interp_spline,
                )

            try:
                self.rotation_3d_z_series = get_inbetweens(
                    parse_key_frames(args.rotation_3d_z),
                    args.max_frames,
                    args.interp_spline,
                )
            except RuntimeError:
                log.warning(
                    "WARNING: You have selected to use key frames, but you have not "
                    "formatted `rotation_3d_z` correctly for key frames.\n"
                    "Attempting to interpret `rotation_3d_z` as "
                    f'"0: ({args.rotation_3d_z})"\n'
                    "Please read the instructions to find out how to use key frames "
                    "correctly.\n"
                )
                args.rotation_3d_z = f"0: ({args.rotation_3d_z})"
                self.rotation_3d_z_series = get_inbetweens(
                    parse_key_frames(args.rotation_3d_z),
                    args.max_frames,
                    args.interp_spline,
                )

        else:
            args.angle = float(args.angle)
            args.zoom = float(args.zoom)
            args.translation_x = float(args.translation_x)
            args.translation_y = float(args.translation_y)
            args.translation_z = float(args.translation_z)
            args.rotation_3d_x = float(args.rotation_3d_x)
            args.rotation_3d_y = float(args.rotation_3d_y)
            args.rotation_3d_z = float(args.rotation_3d_z)

        if eKonf.is_list(args.intermediate_saves):
            args.steps_per_checkpoint = None
        else:
            if (
                not isinstance(args.steps_per_checkpoint, int)
                or args.steps_per_checkpoint < 1
            ):
                if args.intermediate_saves and args.intermediate_saves > 0:
                    args.steps_per_checkpoint = math.floor(
                        (args.steps - args.skip_steps - 1)
                        // (args.intermediate_saves + 1)
                    )
                    args.steps_per_checkpoint = (
                        args.steps_per_checkpoint
                        if args.steps_per_checkpoint > 0
                        else 1
                    )
                else:
                    args.steps_per_checkpoint = None
                    # args.steps_per_checkpoint = args.steps + 10
            log.info(f"Will save every {args.steps_per_checkpoint} steps")
            args.intermediate_saves = []

        timestep_respacing = f"ddim{args.steps}"
        diffusion_steps = (
            (1000 // args.steps) * args.steps if args.steps < 1000 else args.steps
        )
        self._model_and_diffusion_args.update(
            {
                "timestep_respacing": timestep_respacing,
                "diffusion_steps": diffusion_steps,
            }
        )
        skip_step_ratio = int(args.frames_skip_steps.rstrip("%")) / 100
        args.calc_frames_skip_steps = math.floor(args.steps * skip_step_ratio)

        if args.animation_mode == AnimMode.VIDEO_INPUT:
            self.init_flow_warp(args)

            frames = sorted(glob(video_frames_dir + "/*.*"))
            if len(frames) == 0:
                raise Exception(f"ERROR: 0 frames found in {video_frames_dir}.")
            flows = glob(flo_dir + "/*.*")
            if (len(flows) == 0) and args.video_init_flow_warp:
                raise Exception(f"ERROR: 0 flow files found in {flo_dir}.")

        if args.steps <= args.calc_frames_skip_steps:
            raise Exception("ERROR: You can't skip more steps than your total steps")

        batch_config_file = os.path.join(
            batch_dir, f"{args.batch_name}(*)_settings.yaml"
        )
        if args.resume_run:
            if args.batch_num is None:
                if args.run_to_resume == "latest":
                    args.batch_num = len(glob(batch_config_file)) - 1
                    log.info("Resuming latest batch")
                else:
                    args.batch_num = args.run_to_resume
                    log.info("Resuming batch with run_to_resume as batch_num")
            log.info(f"Resuming batch_num: {args.batch_num}")

            _frame_file = os.path.join(
                batch_dir, f"{args.batch_name}({args.batch_num})_*.png"
            )
            if args.resume_from_frame == "latest":
                start_frame = len(glob(_frame_file))
                if (
                    args.animation_mode != AnimMode.ANIM_3D
                    and args.turbo_mode
                    and start_frame > args.turbo_preroll
                    and start_frame % int(args.turbo_steps) != 0
                ):
                    start_frame = start_frame - (start_frame % int(args.turbo_steps))
            else:
                start_frame = int(args.resume_from_frame) + 1
                if (
                    args.animation_mode != AnimMode.ANIM_3D
                    and args.turbo_mode
                    and start_frame > args.turbo_preroll
                    and start_frame % int(args.turbo_steps) != 0
                ):
                    start_frame = start_frame - (start_frame % int(args.turbo_steps))
                if args.retain_overwritten_frames:
                    existing_frames = len(glob(_frame_file))
                    frames_to_save = existing_frames - start_frame
                    log.info(f"Moving {frames_to_save} frames to the Retained folder")
                    move_files(
                        start_frame,
                        existing_frames,
                        self._output.batch_dir,
                        self._output.retain_dir,
                        args.batch_name,
                        args.batch_num,
                    )
        else:
            start_frame = 0
            args.batch_num = len(glob(batch_config_file))
        args.start_frame = start_frame
        args.start_sample = start_frame

        if args.set_seed == "random_seed":
            random.seed()
            args.seed = random.randint(0, 2**32)
        else:
            args.seed = int(args.set_seed)
        log.info(f"Using seed: {args.seed}")

        args.n_samples = args.n_samples if args.animation_mode == AnimMode.NONE else 1
        args.max_frames = args.max_frames if args.animation_mode != AnimMode.NONE else 1

        if args.animation_mode == AnimMode.VIDEO_INPUT:
            # This isn't great in terms of what will get saved to the settings.. but it should work.
            args.clip_guidance_scale = args.video_init_clip_guidance_scale
            args.tv_scale = args.video_init_tv_scale
            args.range_scale = args.video_init_range_scale
            args.sat_scale = args.video_init_sat_scale
            args.cutn_batches = args.video_init_cutn_batches
            args.skip_steps = args.video_init_skip_steps
            args.frames_scale = args.video_init_frames_scale
            args.frames_skip_steps = args.video_init_frames_skip_steps

        if (
            args.models.diffusion_model in self.model_map.kaliyuga_pixel_art_model_names
        ) or (
            args.models.diffusion_model in self.model_map.kaliyuga_pulpscifi_model_names
        ):
            args.cut_overview = args.pad_or_pulp_cut_overview
            args.cut_innercut = args.pad_or_pulp_cut_innercut
            args.cut_ic_pow = args.pad_or_pulp_cut_ic_pow
            args.cut_icgray_p = args.pad_or_pulp_cut_icgray_p
        elif (
            args.models.diffusion_model
            in self.model_map.kaliyuga_watercolor_model_names
        ):
            args.cut_overview = args.watercolor_cut_overview
            args.cut_innercut = args.watercolor_cut_innercut
            args.cut_ic_pow = args.watercolor_cut_ic_pow
            args.cut_icgray_p = args.watercolor_cut_icgray_p

        return args

    def make_gif(
        self,
        batch_name=None,
        batch_num=None,
        sample_num=0,
        output_file=None,
        filename_patterns=None,
        duration=100,
        loop=0,
        width=None,
        optimize=True,
        quality=50,
        show=False,
        force=False,
        **kwargs,
    ):
        args = self.load_config(batch_name, batch_num, **kwargs)
        batch_name = batch_name or args.batch_name
        if batch_num is None:
            batch_num = args.batch_num
        base_dir = self._output.partial_dir

        filename_patterns = (
            filename_patterns or f"{batch_name}({batch_num})_{sample_num:04}-*.png"
        )
        output_file = output_file or f"{batch_name}({batch_num})_{sample_num:04}.gif"
        output_path = os.path.join(self._output.batch_dir, output_file)

        eKonf.make_gif(
            filename_patterns=filename_patterns,
            base_dir=base_dir,
            output_filepath=output_path,
            duration=duration,
            loop=loop,
            width=width,
            optimize=optimize,
            quality=quality,
            show=show,
            force=force,
            **kwargs,
        )

    def _prepare_folders(self, batch_name):
        super()._prepare_folders(batch_name)
        o = self._output
        batch_dir = o.batch_dir
        o.retain_dir = os.path.join(batch_dir, "retained")
        o.partial_dir = os.path.join(batch_dir, "partials")
        o.video_frames_dir = os.path.join(batch_dir, "video_frames")
        o.flo_dir = os.path.join(o.video_frames_dir, "out_flo_fwd")
        o.temp_flo_dir = os.path.join(o.video_frames_dir, "temp_flo")
        o.flo_fwd_dir = os.path.join(o.video_frames_dir, "out_flo_fwd")
        o.flo_out_dir = os.path.join(batch_dir, "flow")
        o.blend_out_dir = os.path.join(batch_dir, "blend")
        for _name, _path in o.items():
            if _name.endswith("_dir") and not os.path.exists(_path):
                os.makedirs(_path)
        o.prev_frame_path = os.path.join(batch_dir, "prev_frame.png")
        o.prev_frame_scaled_path = os.path.join(batch_dir, "prev_frame_scaled.png")
        o.progress_path = os.path.join(batch_dir, "progress.png")
        o.warped_path = os.path.join(batch_dir, "warped.png")
        o.old_frame_scaled_path = os.path.join(batch_dir, "old_frame_scaled.png")

    def create_video(self, batch_name=None, batch_num=None, force=False, **kwargs):
        import PIL
        from tqdm.notebook import trange
        from .flow import warp

        args = self.load_config(batch_name, batch_num)
        batch_name = batch_name or args.batch_name
        if batch_num is None:
            batch_num = args.batch_num
        # @title ### **Create video**
        # @markdown Video file will save in the same folder as your images.

        _video = args.video_output
        _video = eKonf.merge(_video, kwargs)

        flo_dir = self._output.flo_dir
        batch_dir = self._output.batch_dir
        flo_out_dir = self._output.flo_out_dir
        blend_out_dir = self._output.blend_out_dir

        if _video.skip_video_for_run_all:
            print(
                "Skipping video creation, set skip_video_for_run_all=False if you want to run it"
            )

        else:
            frame_files = f"{batch_name}({batch_num})_*.png"
            frames_path = os.path.join(batch_dir, frame_files)
            frames_in = glob(frames_path)

            if _video.last_frame == "final_frame":
                _video.last_frame = len(frames_in)
                log.info(f"Total frames: {_video.last_frame}")

            frame_filename = f"{batch_name}({batch_num})_%04d.png"
            frames_in_path = os.path.join(batch_dir, frame_filename)
            mp4_filename = f"{batch_name}({batch_num}).mp4"

            if (args.video_init_blend_mode == "optical flow") and (
                args.animation_mode == AnimMode.VIDEO_INPUT
            ):
                frames_path = os.path.join(flo_out_dir, frame_files)
                if _video.last_frame == "final_frame":
                    _video.last_frame = len(glob(frames_path))

                frames_in_path = os.path.join(flo_out_dir, frame_filename)
                mp4_filename = f"{batch_name}({batch_num})_flow.mp4"
                shutil.copy(frames_in[0], flo_out_dir)
                log.info(f"Copying first frame to {flo_out_dir}")

                for i in trange(
                    _video.init_frame, min(len(frames_in), _video.last_frame)
                ):
                    frame1_path = frames_in[i - 1]
                    frame2_path = frames_in[i]

                    frame1 = PIL.Image.open(frame1_path)
                    frame2 = PIL.Image.open(frame2_path)
                    frame1_stem = f"{(int(frame1_path.split('/')[-1].split('_')[-1][:-4])+1):04}.jpg"

                    _flo_path = os.path.join(flo_dir, f"{frame1_stem}.npy")
                    log.info(f"Loading flow from {_flo_path}")
                    weights_path = None
                    if _video.video_init_check_consistency:
                        # TBD
                        pass

                    _filename = f"{batch_name}({batch_num})_{i:04}.png"
                    _frame_path = os.path.join(flo_out_dir, _filename)
                    warp(
                        frame1,
                        frame2,
                        _flo_path,
                        blend=_video.blend,
                        weights_path=weights_path,
                    ).save(_frame_path)
                    log.info(f"Saving warped frame {_frame_path}")

            if args.video_init_blend_mode == "linear":
                frames_path = os.path.join(blend_out_dir, frame_files)
                if _video.last_frame == "final_frame":
                    _video.last_frame = len(glob(frames_path))

                frames_in_path = os.path.join(blend_out_dir, frame_filename)
                mp4_filename = f"{batch_name}({batch_num})_blend.mp4"
                shutil.copy(frames_in[0], blend_out_dir)
                log.info(f"Copying first frame to {blend_out_dir}")

                for i in trange(1, len(frames_in)):
                    frame1_path = frames_in[i - 1]
                    frame2_path = frames_in[i]

                    frame1 = PIL.Image.open(frame1_path)
                    frame2 = PIL.Image.open(frame2_path)

                    _filename = f"{batch_name}({batch_num})_{i:04}.png"
                    _frame_path = os.path.join(blend_out_dir, _filename)
                    frame = PIL.Image.fromarray(
                        (
                            np.array(frame1) * (1 - _video.blend)
                            + np.array(frame2) * (_video.blend)
                        ).astype("uint8")
                    ).save(_frame_path)
                    log.info(f"Saving blended frame {_frame_path}")

            mp4_path = os.path.join(batch_dir, mp4_filename)

            return _create_video(
                base_dir=batch_dir,
                video_path=mp4_path,
                input_url=frames_in_path,
                fps=_video.fps,
                start_number=_video.init_frame,
                vframes=(_video.last_frame + 1),
                force=force,
            )

    def parameters(self, name=None):
        if name is None:
            eKonf.print(self._parameters)
        else:
            if name in self._parameters:
                eKonf.print(self._parameters[name])
            else:
                for cfg in self._parameters:
                    for k, v in self._parameters[cfg].items():
                        if re.search(name, k):
                            print(f"[{k}]")
                            print(v)
                            print("")



import _pickle as cPickle
import pickle
import bz2
import sys


def save_pickle(data, filename):
    filename = str(filename)
    with open(filename, "wb") as f:
        pickle.dump(data, f)


def load_pickle(filename):
    filename = str(filename)
    with open(filename, "rb") as f:
        data = pickle.load(f)
    return data


def compressed_pickle(data, filename, add_suffix=False, suffix=".pkl.bz2"):
    """
    save object to file using pickle

    @param filename: name of destination file
    @type filename: str
    @param data: object to save (has to be pickleable)
    @type data: obj
    """
    filename = str(filename)
    if add_suffix and not filename.endswith(suffix):
        filename += suffix

    try:
        f = bz2.BZ2File(filename, "wb")
    except IOError:
        sys.stderr.write("File " + filename + " cannot be written\n")
        return

    cPickle.dump(data, f)
    f.close()


def decompress_pickle(filename):
    """
    Load from filename using pickle

    @param filename: name of file to load from
    @type filename: str
    """

    filename = str(filename)
    try:
        f = bz2.BZ2File(filename, "rb")
    except IOError:
        sys.stderr.write("File " + filename + " cannot be read\n")
        return

    data = cPickle.load(f)
    f.close()
    return data



import logging
import multiprocessing
from contextlib import closing
import scipy.sparse as ssp
import random
import pandas as pd
from math import ceil
from tqdm.auto import tqdm
import contextlib

batcher_instance = None

log = logging.getLogger(__name__)


class Batcher(object):
    """Scheduler to handle parallel jobs on minibatches

    Parameters
    ----------
    procs: int
            Number of process(es)/thread(s) for executing task in parallel. Used for multiprocessing, threading and Loky

    minibatch_size: int
            Expected size of each minibatch

    backend: {'serial', 'multiprocessing', 'threading', 'loky', 'spark', 'dask', 'ray'}
            Backend for computing the tasks

                    - 'serial' sequential execution without a backend scheduler

                    - 'multiprocessing' Python standard multiprocessing library

                    - 'threading' Python standard threading library

                    - 'loky' Loky fork of multiprocessing library

                    - 'joblib' Joblib fork of multiprocessing library

                    - 'spark' PySpark local or distributed execution

                    - 'dask' Dask Distributed local or distributed execution

                    - 'ray' Ray local or distributed execution

    task_num_cpus: int
            Number of CPUs to reserve per minibatch task for Ray

    task_num_gpus: int
            Number of GPUs to reserve per minibatch task for Ray

    backend_handle: object
            Backend handle for sending tasks

    verbose: int
            Verbosity level.
            Setting verbose > 0 will display additional information depending on the specific level set.
    """

    def __init__(
        self,
        procs=0,
        minibatch_size=20000,
        backend_handle=None,
        backend="multiprocessing",
        task_num_cpus=1,
        task_num_gpus=0,
        verbose=0,
    ):
        if isinstance(procs, str):
            procs = int(procs)
        if procs == 0 or procs is None:
            procs = multiprocessing.cpu_count()
        self.procs = procs
        self.verbose = verbose
        self.minibatch_size = minibatch_size
        self.backend_handle = backend_handle
        self.backend = backend
        self.task_num_cpus = task_num_cpus
        self.task_num_gpus = task_num_gpus

    def list2indexedrdd(self, lst, minibatch_size=0):
        if minibatch_size == 0:
            minibatch_size = self.minibatch_size
        start = 0
        len_data = len(lst)
        batch_count = 0
        batches = []
        while start < len_data:
            batches.append([batch_count] + [lst[start : start + minibatch_size]])
            start += minibatch_size
            batch_count += 1
        return self.backend_handle.parallelize(batches)

    def indexedrdd2list(self, indexedrdd, sort=True):
        batches = indexedrdd.collect()
        if sort:
            batches = sorted(batches)
        return [batch[1] for batch in batches]

    def split_batches(self, data, minibatch_size=None, backend=None):
        """Split data into minibatches with a specified size

        Parameters
        ----------
        data: iterable and indexable
                List-like data to be split into batches. Includes backend_handleipy matrices and Pandas DataFrames.

        minibatch_size: int
                Expected sizes of minibatches split from the data.

        backend: object
                Backend to use, instead of the Batcher backend attribute

        Returns
        -------
        data_split: list
                List of minibatches, each entry is a list-like object representing the data subset in a batch.
        """
        if minibatch_size is None:
            minibatch_size = self.minibatch_size
        if backend is None:
            backend = self.backend
        if isinstance(data, list) or isinstance(data, tuple) or isinstance(data, dict):
            len_data = len(data)
        else:
            len_data = data.shape[0]
        if backend == "spark":
            return self.list2indexedrdd(data, minibatch_size)
        if isinstance(data, pd.DataFrame):
            data = [
                data.iloc[x * minibatch_size : (x + 1) * minibatch_size]
                for x in range(int(ceil(len_data / minibatch_size)))
            ]
        elif isinstance(data, dict):
            data = [
                dict(
                    list(data.items())[
                        x * minibatch_size : min(len_data, (x + 1) * minibatch_size)
                    ]
                )
                for x in range(int(ceil(len_data / minibatch_size)))
            ]
        else:
            data = [
                data[x * minibatch_size : min(len_data, (x + 1) * minibatch_size)]
                for x in range(int(ceil(len_data / minibatch_size)))
            ]
        # if backend=="dask":  return self.backend_handle.scatter(data)
        return data

    def collect_batches(self, data, backend=None, sort=True):
        if backend is None:
            backend = self.backend
        if backend == "spark":
            data = self.indexedrdd2list(data, sort)
        if backend == "dask":
            data = self.backend_handle.gather(data)
        return data

    def merge_batches(self, data):
        """Merge a list of data minibatches into one single instance representing the data

        Parameters
        ----------
        data: list
                List of minibatches to merge

        Returns
        -------
        (anonymous): sparse matrix | pd.DataFrame | list
                Single complete list-like data merged from given batches
        """
        if isinstance(data[0], ssp.csr_matrix):
            return ssp.vstack(data)
        if isinstance(data[0], pd.DataFrame) or isinstance(data[0], pd.Series):
            return pd.concat(data)
        return [item for sublist in data for item in sublist]

    def process_batches(
        self,
        task,
        data,
        args,
        backend=None,
        backend_handle=None,
        input_split=False,
        merge_output=True,
        minibatch_size=None,
        procs=None,
        task_num_cpus=None,
        task_num_gpus=None,
        verbose=None,
        description="batch_apply",
    ):
        """

        Parameters
        ----------
        task: function
                Function to apply on each minibatch with other specified arguments

        data: list-like
                Samples to split into minibatches and apply the specified function on

        args: list
                Arguments to pass to the specified function following the mini-batch

        input_split: boolean, default False
                If True, input data is already mapped into minibatches, otherwise data will be split on call.

        merge_output: boolean, default True
                If True, results from minibatches will be reduced into one single instance before return.

        procs: int
                Number of process(es)/thread(s) for executing task in parallel. Used for multiprocessing, threading,
                Loky and Ray

        minibatch_size: int
                Expected size of each minibatch

        backend: {'serial', 'multiprocessing', 'threading', 'loky', 'spark', 'dask', 'ray'}
                Backend for computing the tasks

                        - 'serial' sequential execution without a backend scheduler

                        - 'multiprocessing' Python standard multiprocessing library

                        - 'threading' Python standard threading library

                        - 'loky' Loky fork of multiprocessing library

                        - 'joblib' Joblib fork of multiprocessing library

                        - 'spark' PySpark local or distributed execution

                        - 'dask' Dask Distributed local or distributed execution

                        - 'ray' Ray local or distributed execution

        backend_handle: object
                Backend handle for sending tasks

        task_num_cpus: int
                Number of CPUs to reserve per minibatch task for Ray

        task_num_gpus: int
                Number of GPUs to reserve per minibatch task for Ray

        verbose: int
                Verbosity level.
                Setting verbose > 0 will display additional information depending on the specific level set.

        Returns
        -------
        results: list-like | list of list-like
                If merge_output is specified as True, this will be a list-like object representing
                the dataset, with each entry as a sample. Otherwise this will be a list of list-like
                objects, with each entry representing the results from a minibatch.
        """
        if procs is None:
            procs = self.procs
        if backend is None:
            backend = self.backend
        if backend_handle is None:
            backend_handle = self.backend_handle
        if task_num_cpus is None:
            task_num_cpus = self.task_num_cpus
        if task_num_gpus is None:
            task_num_gpus = self.task_num_gpus
        if verbose is None:
            verbose = self.verbose
        if verbose > 1:
            log.info(
                "%s %s %s %s %s %s %s %s %s %s %s %s %s %s"
                % (
                    " backend:",
                    backend,
                    " minibatch_size:",
                    self.minibatch_size,
                    " procs:",
                    procs,
                    " input_split:",
                    input_split,
                    " merge_output:",
                    merge_output,
                    " len(data):",
                    len(data),
                    "len(args):",
                    len(args),
                )
            )

        if verbose > 10:
            log.info(
                "%s %s %s %s %s %s %s %s"
                % (
                    " len(data):",
                    len(data),
                    "len(args):",
                    len(args),
                    "[type(x) for x in data]:",
                    [type(x) for x in data],
                    "[type(x) for x in args]:",
                    [type(x) for x in args],
                )
            )

        if not (input_split):
            if backend == "spark":
                paral_params = self.split_batches(data, minibatch_size, backend="spark")
            else:
                paral_params = [
                    [data_batch] + args
                    for data_batch in self.split_batches(data, minibatch_size)
                ]
        else:
            if backend != "spark":
                paral_params = [[data_batch] + args for data_batch in data]
            else:
                paral_params = data
        if verbose > 10:
            print(" Start task, len(paral_params)", len(paral_params))
        if backend == "serial":
            results = [
                task(minibatch) for minibatch in tqdm(paral_params, desc=description)
            ]
        else:
            if backend == "multiprocessing":
                with closing(
                    multiprocessing.Pool(max(1, procs), maxtasksperchild=2)
                ) as pool:
                    results = pool.map_async(task, paral_params)
                    pool.close()
                    pool.join()
                    results = results.get()
            elif backend == "threading":
                with closing(multiprocessing.dummy.Pool(max(1, procs))) as pool:
                    results = pool.map(task, paral_params)
                    pool.close()
                    pool.join()
            elif backend == "loky":
                from loky import get_reusable_executor

                pool = get_reusable_executor(max_workers=max(1, procs))
                results = list(pool.map(task, tqdm(paral_params, desc=description)))
            elif backend == "joblib":
                from joblib import Parallel, delayed

                with tqdm_joblib(
                    tqdm(desc=description, total=len(paral_params))
                ) as pbar:
                    results = Parallel(n_jobs=procs)(
                        delayed(task)(params) for params in paral_params
                    )
            elif backend == "p_tqdm":
                from p_tqdm import p_map

                results = p_map(task, paral_params, num_cpus=procs)
            elif backend == "dask":
                # if not (input_split):  data= self.scatter(data)
                results = [
                    self.backend_handle.submit(task, params)
                    for params in tqdm(paral_params, desc=description)
                ]
            elif backend == "spark":

                def apply_func_to_indexedrdd(batch):
                    return [batch[0]] + [task([batch[1]] + args)]

                results = paral_params.map(apply_func_to_indexedrdd)
            elif backend == "ray":

                @self.backend_handle.remote(
                    num_cpus=task_num_cpus, num_gpus=task_num_gpus
                )
                def f_ray(f, data):
                    return f(data)

                pbar = tqdm(desc=description, total=len(paral_params) + 1)
                results = [
                    f_ray.remote(task, paral_params.pop(0))
                    for _ in range(min(len(paral_params), self.procs))
                ]
                uncompleted = results
                pbar.update(len(results))
                while len(paral_params) > 0:
                    # More tasks than available processors. Queue the task calls
                    done, remaining = self.backend_handle.wait(
                        uncompleted, timeout=60, fetch_local=False
                    )
                    # if verbose > 5:  print("Done, len(done), len(remaining)", len(done), len(remaining))
                    if len(done) == 0:
                        continue
                    done = done[0]
                    uncompleted = [x for x in uncompleted if x != done]
                    if len(remaining) > 0:
                        new = f_ray.remote(task, paral_params.pop(0))
                        pbar.update(1)
                        uncompleted.append(new)
                        results.append(new)
                results = [self.backend_handle.get(x) for x in results]
                pbar.update(1)
                pbar.close()
            # ppft currently not supported. Supporting arbitrary tasks requires modifications to passed arguments
            # elif backend == "ppft":
            #   jobs = [self.backend_handle.submit(task, (x,), (), ()) for x in paral_params]
            # 	results = [x() for x in jobs]

        if merge_output:
            return self.merge_batches(self.collect_batches(results, backend=backend))
        if verbose > 2:
            log.info(
                "%s %s %s %s %s %s %s"
                % (
                    " Task:",
                    task,
                    " backend:",
                    backend,
                    " backend_handle:",
                    backend_handle,
                    " completed",
                )
            )
        return results

    def shuffle_batch(self, texts, labels=None, seed=None):
        """Shuffle a list of samples, as well as the labels if specified

        Parameters
        ----------
        texts: list-like
                List of samples to shuffle

        labels: list-like (optional)
                List of labels to shuffle, should be correspondent to the samples given

        seed: int
                The seed of the pseudo random number generator to use for shuffling

        Returns
        -------
        texts: list
                List of shuffled samples (texts parameters)

        labels: list (optional)
                List of shuffled labels. This will only be returned when non-None
                labels is passed
        """
        if seed is not None:
            random.seed(seed)
        index_shuf = list(range(len(texts)))
        random.shuffle(index_shuf)
        texts = [texts[x] for x in index_shuf]
        if labels is None:
            return texts
        labels = [labels[x] for x in index_shuf]
        return texts, labels

    def __getstate__(self):
        return dict((k, v) for (k, v) in self.__dict__.items())

    def __setstate__(self, params):
        for key in params:
            setattr(self, key, params[key])


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    import joblib

    class TqdmBatchCompletionCallBack(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallBack
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


# from joblib import Parallel, delayed

# with tqdm_joblib(tqdm(desc="My calculation", total=10)) as progress_bar:
#     Parallel(n_jobs=16)(delayed(sqrt)(i**2) for i in range(10))



import logging
from io import StringIO
from zipfile import ZipFile
import pandas as pd
from tqdm.auto import tqdm
from ekorpkit.io.parse.json import parse_data
from joblib import Parallel, delayed
from ekorpkit.utils.batch.batcher import tqdm_joblib
from ekorpkit.io.file import get_files_from_archive, get_filepaths
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def load_data(split_name=None, **loader_cfg):
    data_dir = loader_cfg["data_dir"]
    if split_name:
        data_files = loader_cfg["data_sources"][split_name]
    else:
        data_files = loader_cfg["data_sources"]
    filepaths = get_filepaths(data_files, data_dir)

    num_workers = loader_cfg.get("num_workers", 1)
    filetype = loader_cfg.get("filetype", None)
    # data_items =  args['data']['item'].keys()
    multiprocessing_at = loader_cfg.get("multiprocessing_at", "load_data")

    if split_name:
        default_items = {eKonf.Keys.SPLIT.value: split_name}
    else:
        default_items = {}
    documents = []
    num_workers = num_workers if num_workers else 1
    num_files = len(filepaths)
    if num_files < 1:
        log.warning("No files to load")
        df = pd.DataFrame(documents)
        return df

    processes = min(num_workers, num_files)
    if multiprocessing_at == "load_data" and num_files < processes // 2:
        multiprocessing_at = "_load_archive"
        loader_cfg["multiprocessing_at"] = multiprocessing_at
    if processes > 2 and multiprocessing_at == "load_data":
        log.info(
            f"Starting multiprocessing with {processes} processes at {multiprocessing_at}"
        )
        desciption = "::load_data()"
        with tqdm_joblib(tqdm(desc=desciption, total=num_files)):
            results = Parallel(n_jobs=num_workers)(
                delayed(_load_archive)(filepath, filetype, default_items, loader_cfg, 1)
                for filepath in filepaths
            )
            for result in results:
                documents += result
    else:
        for fno, filepath in enumerate(filepaths):
            log.info(f"==> processing {(fno+1)}/{num_files} files <==")
            documents += _load_archive(
                filepath, filetype, default_items, loader_cfg, num_workers
            )

    print(documents[0])
    df = pd.DataFrame(documents)
    return df


def _load_archive(filepath, filetype, default_items, loader_args, num_workers):
    from loky import get_reusable_executor

    multiprocessing_at = loader_args.get("multiprocessing_at", "load_data")
    limit = loader_args.get("limit", None)

    files, arch_handle, open_func = get_files_from_archive(filepath, filetype=filetype)
    num_files = len(files)
    if (
        num_workers > 2
        and multiprocessing_at == "_load_archive"
        and num_files > num_workers
    ):
        log.info(
            f"Starting multiprocessing with {num_workers} processes at {multiprocessing_at}"
        )
        desciption = "::_load_archive()"
        executor = get_reusable_executor(max_workers=num_workers)
        pbar = tqdm(total=num_files, desc=desciption)
    else:
        executor = None

    documents = []
    results = []
    for fno, (file, filename) in enumerate(files):
        if limit and fno >= limit:
            if len(documents) > 0:
                print(documents[0])
            break
        # default_items["filename"] = filename

        try:
            if open_func is not None:
                with open_func(file) as f:
                    contents = f.read()
            else:
                with open(file, "rb") as f:
                    contents = f.read()
        except Exception as e:
            log.critical(f"Error reading {filename}", e)
            continue

        if contents is None:
            log.warning("contents is empty, skipping")
            continue
        if executor is not None:
            pbar.set_description(f"{filename}")
            results.append(
                executor.submit(parse_data, contents, loader_args, default_items, filename, 1)
            )
            pbar.update(1)
        else:
            documents += parse_data(contents, loader_args, default_items, filename, num_workers)

    if arch_handle is not None:
        arch_handle.close()
    if executor is not None:
        # pbar.close()
        for result in results:
            documents += result.result()
    return documents


def load_dataframe(split_name, **loader_cfg):
    data_dir = loader_cfg["data_dir"]
    data_files = loader_cfg["data_sources"][split_name]
    filepaths = get_filepaths(data_files, data_dir)

    documents = []
    log.info(f"Loading [{split_name}] documents from {len(filepaths)} files")
    for filepath in filepaths:
        log.info("==> processing {}".format(filepath))
        if filepath.endswith(".zip"):
            with ZipFile(filepath) as zf:
                for f_csv in zf.namelist():
                    with zf.open(f_csv, "r") as zfo:
                        content = zfo.read().decode(errors="ignore")
            df = pd.read_csv(StringIO(content), index_col=None)
        else:
            df = eKonf.load_data(filepath, index_col=None)
        documents.append(df)
    df = pd.concat(documents, ignore_index=True)
    return df


def load_hfds(split_name, **loader_cfg):
    from datasets import load_dataset

    dataset_name = loader_cfg["name"]
    split = loader_cfg["data_sources"][split_name]
    subsets = loader_cfg["subset"]
    download_mode = loader_cfg.get("download_mode", "force_redownload")
    ignore_verifications = loader_cfg.get("ignore_verifications", True)

    if isinstance(subsets, str):
        subsets = [subsets]
    elif not isinstance(subsets, list):
        subsets = [None]

    log.info(
        f"Loading [{split_name}] documents from huggingface datasets [{dataset_name}]"
    )

    dfs = []
    for subset in subsets:
        ds = load_dataset(
            dataset_name,
            subset,
            split=split,
            download_mode=download_mode,
            ignore_verifications=ignore_verifications,
        )
        print(ds)
        df = ds.to_pandas()
        df["subset"] = subset
        dfs.append(df)
    df = pd.concat(dfs, ignore_index=True)
    return df


def load_tsv_data(split_name, **loader_cfg):
    data_dir = loader_cfg["data_dir"]
    data_files = loader_cfg["data_sources"][split_name]
    filepaths = get_filepaths(data_files, data_dir)
    data_args = loader_cfg["data"]
    sep = data_args.get("sep", "\t")
    sep = str(sep).encode("utf-8").decode("unicode_escape")
    dtype = dict(data_args.get("dtype", None))
    names = list(dtype.keys())
    constants = data_args.get("constants", None)

    documents = []
    log.info(f"Loading [{split_name}] documents from {len(filepaths)} files")
    for filepath in filepaths:
        log.info("==> processing {}".format(filepath))
        if filepath.endswith(".zip"):
            with ZipFile(filepath) as zf:
                for f_csv in zf.namelist():
                    with zf.open(f_csv, "r") as zfo:
                        content = zfo.read().decode(errors="ignore")
            df = pd.read_csv(
                StringIO(content), index_col=None, sep=sep, dtype=dtype, names=names
            )
        else:
            df = pd.read_csv(
                filepath, index_col=None, sep=sep, dtype=dtype, names=names
            )
        documents.append(df)
    df = pd.concat(documents, ignore_index=True)
    if constants is not None:
        for k, v in constants.items():
            df[k] = v
    return df


def read_conll(filename, **args):
    sep = "\t" if args["sep"] is None else args["sep"]
    column_dtypes = args["column_dtypes"]
    names = (
        ["words", "pos", "chunk", "labels"]
        if column_dtypes is None
        else column_dtypes.keys()
    )
    filter_query = args["filter_query"]

    df = pd.read_csv(
        filename,
        sep=sep,
        names=names,
        header=None,
        keep_default_na=False,
        quoting=3,
        skip_blank_lines=False,
        engine="python",
    )
    df = df[
        ~df["words"].astype(str).str.startswith("-DOCSTART-")
    ]  # Remove the -DOCSTART- header
    df["sentence_id"] = (df.words == "").cumsum()
    df = df[df.words != ""]
    df = df.astype(dict(column_dtypes), errors="raise")
    if filter_query:
        df = df.query(filter_query, engine="python")
    return df


def read_csv(filename, **args):
    sep = "," if args["sep"] is None else args["sep"]
    column_mapping = args["column_mapping"] if "column_mapping" in args else None
    filter_query = args["filter_query"] if "filter_query" in args else None
    index_col = args["index_col"] if "index_col" in args else None

    df = pd.read_csv(filename, sep=sep, index_col=index_col, engine="python")
    if column_mapping is not None:
        df.rename(columns=column_mapping, inplace=True)
    if filter_query:
        df = df.query(filter_query, engine="python")
    return df



from typing import Iterator, List

from tqdm import tqdm

from ..policy.core import Policy
from ..tf import BaseTransformationFunction
from ...types import DataPoint, DataPoints
from ...utils.data_operators import check_unique_names


class BaseTFApplier:
    """Base class for TF applier objects.

    Base class for TF applier objects, which execute a set of TF
    on a collection of data points. Subclasses should operate on
    a single data point collection format (e.g. ``DataFrame``).
    Subclasses must implement the ``apply`` method.

    Parameters
    ----------
    tfs
        TFs that this applier executes on examples
    policy
        Augmentation policy used to generate sequences of TFs

    Raises
    ------
    ValueError
        If names of TFs are not unique
    """

    def __init__(self, tfs: List[BaseTransformationFunction], policy: Policy) -> None:
        self._tfs = tfs
        self._tf_names = [tf.name for tf in tfs]
        check_unique_names(self._tf_names)
        self._policy = policy

    def _apply_policy_to_data_point(self, x: DataPoint) -> DataPoints:
        x_transformed = []
        for seq in self._policy.generate_for_example():
            x_t = x
            # Handle empty sequence for `keep_original`
            transform_applied = len(seq) == 0
            # Apply TFs
            for tf_idx in seq:
                tf = self._tfs[tf_idx]
                x_t_or_none = tf(x_t)
                # Update if transformation was applied
                if x_t_or_none is not None:
                    transform_applied = True
                    x_t = x_t_or_none
            # Add example if original or transformations applied
            if transform_applied:
                x_transformed.append(x_t)
        return x_transformed

    def __repr__(self) -> str:
        policy_name = type(self._policy).__name__
        return f"{type(self).__name__}, Policy: {policy_name}, TFs: {self._tf_names}"


class TFApplier(BaseTFApplier):
    """TF applier for a list of data points.

    Augments a list of data points (e.g. ``SimpleNamespace``). Primarily
    useful for testing.
    """

    def apply_generator(
        self, data_points: DataPoints, batch_size: int
    ) -> Iterator[List[DataPoint]]:
        """Augment a list of data points using TFs and policy in batches.

        This method acts as a generator, yielding augmented data points for
        a given input batch of data points. This can be useful in a training
        loop when it is too memory-intensive to pregenerate all transformed
        examples.

        Parameters
        ----------
        data_points
            List containing data points to be transformed
        batch_size
            Batch size for generator. Yields augmented data points
            for the next ``batch_size`` input data points.

        Yields
        ------
        List[DataPoint]
            List of data points in augmented data set for batches of inputs
        """
        for i in range(0, len(data_points), batch_size):
            batch_transformed: List[DataPoint] = []
            for x in data_points[i : i + batch_size]:
                batch_transformed.extend(self._apply_policy_to_data_point(x))
            yield batch_transformed

    def apply(
        self, data_points: DataPoints, progress_bar: bool = True
    ) -> List[DataPoint]:
        """Augment a list of data points using TFs and policy.

        Parameters
        ----------
        data_points
            List containing data points to be transformed
        progress_bar
            Display a progress bar?

        Returns
        -------
        List[DataPoint]
            List of data points in augmented data set
        """
        x_transformed: List[DataPoint] = []
        for x in tqdm(data_points, disable=(not progress_bar)):
            x_transformed.extend(self._apply_policy_to_data_point(x))
        return x_transformed



import os
from ekorpkit.base import _getLogger, _env_set


logger = _getLogger(__name__)


def _mount_google_drive(
    workspace=None,
    project=None,
    mountpoint="/content/drive",
    force_remount=False,
    timeout_ms=120000,
):
    try:
        from google.colab import drive

        drive.mount(mountpoint, force_remount=force_remount, timeout_ms=timeout_ms)

        if isinstance(workspace, str):
            if not workspace.startswith(os.path.sep) and not workspace.startswith(".."):
                workspace = os.path.join(mountpoint, workspace)
            _env_set("EKORPKIT_WORKSPACE_ROOT", workspace)
            logger.info(f"Setting EKORPKIT_WORKSPACE_ROOT to {workspace}")
        if isinstance(project, str):
            _env_set("EKORPKIT_PROJECT", project)
            logger.info(f"Setting EKORPKIT_PROJECT to {project}")
    except ImportError:
        logger.warning("Google Colab not detected.")



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>
# Original code from https://github.com/andrei-dubovik/repec
# Modified to return list of dictionaries instead of tuples
"""Routines for decoding and parsing ReDIF documents."""

# Load global packages
import re
from collections import defaultdict

# Define constants
CLUSTERS = set(
    [
        "author-name",
        "editor-name",
        "file-url",
        "provider-name",
        "workplace-name",
    ]
)


def decode(rdf, hint=[]):
    """Decode ReDIF document."""

    def decode(encoding):
        rslt = rdf.decode(encoding)
        if rslt.lower().find("template-type") == -1:
            raise RuntimeError("Decoding Error")
        return rslt

    encodings = hint + ["windows-1252", "utf-8", "utf-16", "latin-1"]
    if rdf[:3] == b"\xef\xbb\xbf":
        encodings = ["utf-8-sig"] + encodings
    for enc in encodings:
        try:
            return decode(enc)
        except Exception:
            continue
    raise RuntimeError("Decoding Error")


def split(lst, sel):
    """Split a list using a selector function."""
    group = {}
    groups = [group]
    for el in lst:
        if sel(el):
            group = {}
            groups.append(group)
        group[el[0]] = el[1]
    return groups


def load(rdf):
    """Load ReDIF document."""
    # Repair line endings
    rdf = re.sub("\r(?!\n)", "\r\n", rdf, flags=re.M)

    # Drop comments
    rdf = re.sub("^#.*\n?", "", rdf, flags=re.M)

    # Split fields
    rdf = re.split(r"(^[a-zA-Z0-9\-#]+\s*:\s*)", rdf, flags=re.M)[1:]
    rdf = [line.strip() for line in rdf]
    rdf = [
        (rdf[i].rstrip(":").rstrip().lower(), rdf[i + 1]) for i in range(0, len(rdf), 2)
    ]

    # Drop empty fields
    rdf = [f for f in rdf if f[1] != ""]

    # Split templates
    rdf = split(rdf, lambda x: x[0] == "template-type")[1:]
    return rdf


def collect(records):
    """Collect ReDIF fields together, group clusters."""

    def helper(path, head, doc, i):
        flag = False
        while i < len(records):
            k, v = records[i]
            if k.startswith(path):
                k = k[len(path) :]
                if k in CLUSTERS:
                    # Enter new cluster
                    base, _, key = k.partition("-")
                    cluster = defaultdict(lambda: [])
                    doc[base].append(cluster)
                    i = helper(path + base + "-", key, cluster, i)
                elif flag and k == head:
                    # Exit cluster on a repeated cluster key
                    return i
                else:
                    doc[k].append(v)
                    flag = True
                    i += 1
            else:
                # Exit cluster
                return i
        return i

    doc = defaultdict(lambda: [])
    helper("", "", doc, 0)
    return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2021-07-24 11:30:23
# @Author  : Chenghao Mou (mouchenghao@gmail.com)

"""Deduplicating Training Data Makes Language Models Better."""

import numpy as np
import multiprocessing as mp
from typing import List, Any, Tuple
from multiprocessing import Manager
from ctypes import c_char_p
from numpy.lib.stride_tricks import sliding_window_view
from nltk.util import ngrams


class MinHashDeduper:
    def __init__(
        self, num_perm: int = 128, threshold: float = 0.5, ngram_size: int = 5
    ):

        self.num_perm = num_perm
        self.threshold = threshold
        self.ngram_size = ngram_size
        self.lsh = None

    def fit_transform(self, data: List[str]) -> List[int]:
        from datasketch import MinHash, MinHashLSH

        """Group similar documents with minhash.

        Parameters
        ----------
        data : List[str]
            List of document strings.

        Returns
        -------
        List[int]
            List of group indices.
        
        Examples
        --------
        >>> deduper = MinHashDeduper(ngram_size=5, threshold=0.3)
        >>> groups = deduper.fit_transform(["This is a sentence.", "This is another sentence.", "This is a question.", "hello world"])
        >>> groups
        [0, 0, 2, 3]
        """
        self.lsh = MinHashLSH(threshold=self.threshold, num_perm=self.num_perm)
        signatures = []
        for i, doc in enumerate(data):
            m = MinHash(num_perm=self.num_perm)
            for ngram in ngrams(doc, self.ngram_size):
                m.update("".join(ngram).encode("utf-8"))
            signatures.append(m)
            self.lsh.insert(f"m{i}", m)

        neighbors = []
        for i, doc in enumerate(data):
            result = self.lsh.query(signatures[i])
            neighbors.append([int(x[1:]) for x in result])

        return get_group_indices(neighbors)


class SuffixArray:
    def __init__(self, k: int = 50):
        self.array = []
        self.k = k

    def fit_transform(self, data: List[str]) -> Tuple[List[str], np.ndarray]:
        """Find duplicate substrings in the data.

        Parameters
        ----------
        data : List[str]
            List of documents.

        Returns
        -------
        Tuple[List[str], np.ndarray]
            List of duplicate substrings and a matrix where each row is a document and each column is a substring.

        Examples
        --------
        >>> array = SuffixArray(k = 9)
        >>> duplicates, groups = array.fit_transform(["This is a sentence.", "This is another sentences.", "This is a question.", "hello world"] * 10)
        >>> assert len(duplicates) == groups.shape[1], "Invalid number of columns"
        >>> assert groups.shape[0] == 40, "Invalid number of rows"
        """
        S = "".join(data)
        suffixes = []
        for i in range(len(S)):
            suffixes.append(S[i:])

        self.array = np.argsort(suffixes)

        # Find duplicated substrings
        manager = Manager()
        shared = manager.Value(c_char_p, S)

        with mp.Pool(mp.cpu_count()) as pool:
            results = pool.starmap(
                similar,
                [(x, y, shared, self.k) for x, y in sliding_window_view(self.array, 2)],
            )

        duplicates = []
        for idx, dup in zip(self.array, results):
            if dup:
                duplicates.append(S[idx : idx + self.k])

        # Find duplicated documents
        try:
            from multiprocessing import shared_memory

            shared = shared_memory.ShareableList(duplicates)
        except ImportError as e:
            print(
                f"The following error was: \n{e}\n\n"
                + "This was likely raised since you are not running python 3.8 or higher."
                + " Continuing without a shared memory file which is likely be inefficient."
            )
            shared = duplicates
        with mp.Pool(mp.cpu_count()) as pool:
            results = pool.starmap(group, [(d, shared) for d in data])

        shared.shm.close()
        shared.shm.unlink()
        del shared

        groups = np.zeros((len(data), len(duplicates)), dtype=bool)
        for i, x in enumerate(results):
            for y in x:
                groups[i, y] = 1

        return duplicates, groups


def similar(x: int, y: int, S: Any, k: int) -> bool:
    """Whether S[x:x+k] is the same as S[y:y+k].

    Parameters
    ----------
    x : int
        [description]
    y : int
        [description]
    S : Any
        [description]
    k : int
        [description]

    Returns
    -------
    bool
        [description]
    """
    if x == y:
        return True

    return (
        x + k <= len(S.value)
        and y + k <= len(S.value)
        and S.value[x : x + k] == S.value[y : y + k]
    )


def group(x: str, patterns: str) -> List[int]:
    """Find patterns that are present in string x.

    Parameters
    ----------
    x : str
        A document string
    patterns : str
        Patterns to search for

    Returns
    -------
    List[int]
        List of indices of which patterns are present in string x
    """
    result = []
    for idx, pattern in enumerate(patterns):
        if pattern in x:
            result.append(idx)
    return result


def get_group_indices(neighbors: List[List[int]]) -> List[int]:
    """Based on the nearest neighbors, find the group/cluster index for each element.

    Parameters
    ----------
    neighbors : List[List[int]]
        List of nearest neighbor indices

    Returns
    -------
    List[int]
        List of group indices
    """
    finder = UF(len(neighbors))
    for i, n in enumerate(neighbors):
        for j in n:
            finder.union(i, j)

    return [finder.find(i) for i in range(len(neighbors))]


"""This module implements an union find or disjoint set data structure.

Source: https://python-algorithms.readthedocs.io/en/stable/_modules/python_algorithms/basic/union_find.html

An union find data structure can keep track of a set of elements into a number
of disjoint (nonoverlapping) subsets. That is why it is also known as the
disjoint set data structure. Mainly two useful operations on such a data
structure can be performed. A *find* operation determines which subset a
particular element is in. This can be used for determining if two
elements are in the same subset. An *union* Join two subsets into a
single subset.

The complexity of these two operations depend on the particular implementation.
It is possible to achieve constant time (O(1)) for any one of those operations
while the operation is penalized. A balance between the complexities of these
two operations is desirable and achievable following two enhancements:

1.  Using union by rank -- always attach the smaller tree to the root of the
    larger tree.
2.  Using path compression -- flattening the structure of the tree whenever
    find is used on it.

complexity:
    * find -- :math:`O(\\alpha(N))` where :math:`\\alpha(n)` is
      `inverse ackerman function
      <http://en.wikipedia.org/wiki/Ackermann_function#Inverse>`_.
    * union -- :math:`O(\\alpha(N))` where :math:`\\alpha(n)` is
      `inverse ackerman function
      <http://en.wikipedia.org/wiki/Ackermann_function#Inverse>`_.

"""


class UF:  # pragma: no cover
    """An implementation of union find data structure.
    It uses weighted quick union by rank with path compression.
    """

    def __init__(self, N):
        """Initialize an empty union find object with N items.

        Args:
            N: Number of items in the union find object.
        """

        self._id = list(range(N))
        self._count = N
        self._rank = [0] * N

    def find(self, p):
        """Find the set identifier for the item p."""

        id = self._id
        while p != id[p]:
            p = id[p] = id[id[p]]  # Path compression using halving.
        return p

    def count(self):
        """Return the number of items."""

        return self._count

    def connected(self, p, q):
        """Check if the items p and q are on the same set or not."""

        return self.find(p) == self.find(q)

    def union(self, p, q):
        """Combine sets containing p and q into a single set."""

        id = self._id
        rank = self._rank

        i = self.find(p)
        j = self.find(q)
        if i == j:
            return

        self._count -= 1
        if rank[i] < rank[j]:
            id[i] = j
        elif rank[i] > rank[j]:
            id[j] = i
        else:
            id[j] = i
            rank[i] += 1

    def __str__(self):
        """String representation of the union find object."""
        return " ".join([str(x) for x in self._id])

    def __repr__(self):
        """Representation of the union find object."""
        return "UF(" + str(self) + ")"



import random
from typing import Sequence

from snorkel.classification.data import DictDataLoader

from .scheduler import BatchIterator, Scheduler


class ShuffledScheduler(Scheduler):
    """Return batches from all dataloaders in shuffled order for each epoch."""

    def __init__(self) -> None:
        super().__init__()

    def get_batches(self, dataloaders: Sequence[DictDataLoader]) -> BatchIterator:
        """Return batches in shuffled order from dataloaders.

        Note that this shuffles the batch order, but it does not shuffle the datasets
        themselves; shuffling the datasets is specified in the DataLoaders directly.

        Parameters
        ----------
        dataloaders
            A sequence of dataloaders to get batches from

        Yields
        ------
        (batch, dataloader)
            batch is a tuple of (X_dict, Y_dict) and dataloader is the dataloader
            that that batch came from. That dataloader will not be accessed by the
            model; it is passed primarily so that the model can pull the necessary
            metadata to know what to do with the batch it has been given.
        """
        batch_counts = [len(dl) for dl in dataloaders]
        dataloader_iters = [iter(dl) for dl in dataloaders]

        dataloader_indices = []
        for idx, count in enumerate(batch_counts):
            dataloader_indices.extend([idx] * count)

        random.shuffle(dataloader_indices)

        for index in dataloader_indices:
            yield next(dataloader_iters[index]), dataloaders[index]



from ekorpkit import eKonf
from ekorpkit.datasets.corpus import Corpus
from ekorpkit.datasets.corpora import Corpora
from ekorpkit.datasets.dataset import Dataset
from ekorpkit.datasets.datasets import Datasets


class Data:
    def __init__(self, data=None, **args):

        args = eKonf.to_config(args)
        self.args = args
        self._data = None
        self._column = None
        self._name = None
        self.verbose = args.get(eKonf.Keys.VERBOSE, False)

        if eKonf.is_dataframe(data):
            self._data = data
        elif eKonf.is_config(args):
            self.load(**args)

        self._column_info = self.args.column_info
        if self._column_info is None:
            raise ValueError("Column info can't be None")
        if self._column is None:
            self._column = eKonf.instantiate(self._column_info)

    def load(self, **args):
        _corpus = args.get(eKonf.Keys.CORPUS) or {}
        _dataset = args.get(eKonf.Keys.DATASET) or {}
        _path = args.get(eKonf.Keys.PATH) or {}

        if _corpus.get(eKonf.Keys.NAME) and _dataset.get(eKonf.Keys.NAME) is None:
            args = _corpus
        elif _dataset.get(eKonf.Keys.NAME):
            args = _dataset

        if eKonf.is_instantiatable(args):
            data = eKonf.instantiate(args)
            if isinstance(data, (Corpus)):
                self._data = data.data
                self._column = data.COLUMN
            elif isinstance(data, (Corpora)):
                data.concat_corpora()
                self._data = data.data
                self._column = data.COLUMN
            elif isinstance(data, (Dataset, Datasets)):
                self._data = data.splits
                self._column = data.COLUMN
            self._name = data.name
        elif _path:
            self._data = eKonf.load_data(**_path[eKonf.Keys.DATA])
            self._name = _path[eKonf.Keys.DATA][eKonf.Keys.FILE]

    def __str__(self):
        classname = self.__class__.__name__
        return f"{classname}: {self._name}{type(self._data)}"

    @property
    def COLUMN(self):
        return self._column

    @property
    def ID(self):
        return self.COLUMN.ID

    @property
    def IDs(self):
        return self.COLUMN.IDs

    @property
    def TEXT(self):
        return self.COLUMN.TEXT

    @property
    def data(self):
        return self._data



from typing import Tuple

import numpy as np


def generate_simple_label_matrix(
    n: int, m: int, cardinality: int, abstain_multiplier: float = 1.0
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generate a synthetic label matrix with true parameters and labels.

    This function generates a set of labeling function conditional probability tables,
    P(LF=l | Y=y), stored as a matrix P, and true labels Y, and then generates the
    resulting label matrix L.

    Parameters
    ----------
    n
        Number of data points
    m
        Number of labeling functions
    cardinality
        Cardinality of true labels (i.e. not including abstains)
    abstain_multiplier
        Factor to multiply the probability of abstaining by

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        A tuple containing the LF conditional probabilities P,
        the true labels Y, and the output label matrix L
    """
    # Generate the conditional probability tables for the LFs
    # The first axis is LF, second is LF output label, third is true class label
    # Note that we include abstains in the LF output space, and that we bias the
    # conditional probabilities towards being non-adversarial
    P = np.empty((m, cardinality + 1, cardinality))
    for i in range(m):
        p = np.random.rand(cardinality + 1, cardinality)

        # Bias the LFs to being non-adversarial
        p[1:, :] += (cardinality - 1) * np.eye(cardinality)

        # Optionally increase the abstain probability by some multiplier; note this is
        # to simulate the common setting where LFs label very sparsely
        p[0, :] *= abstain_multiplier

        # Normalize the conditional probabilities table
        P[i] = p @ np.diag(1 / p.sum(axis=0))

    # Generate the true datapoint labels
    # Note: Assuming balanced classes to start
    Y = np.random.choice(cardinality, n)

    # Generate the label matrix L
    L: np.ndarray = np.empty((n, m), dtype=int)
    for i in range(n):
        for j in range(m):
            L[i, j] = np.random.choice(cardinality + 1, p=P[j, :, Y[i]]) - 1
    return P, Y, L



from ..map import BaseMapper, LambdaMapper, Mapper, lambda_mapper

"""Base classes for transformation functions.

A transformation function (TF) represents an atomic transformation
to a data point in a data augmentation pipeline. Common examples in
image processing include small image rotations or crops. Snorkel
models data augmentation as a sequence of TFs generated by a policy.
"""

# Used for type checking only
# Note: subclassing as below trips up mypy
BaseTransformationFunction = BaseMapper


class TransformationFunction(Mapper):
    """Base class for TFs.

    See ``snorkel.map.core.Mapper`` for details.
    """

    pass


class LambdaTransformationFunction(LambdaMapper):
    """Convenience class for definining TFs from functions.

    See ``snorkel.map.core.LambdaMapper`` for details.
    """

    pass


class transformation_function(lambda_mapper):
    """Decorate functions to create TFs.

    See ``snorkel.map.core.lambda_mapper`` for details.

    Example
    -------
    >>> @transformation_function()
    ... def square(x):
    ...     x.num = x.num ** 2
    ...     return x
    >>> from types import SimpleNamespace
    >>> square(SimpleNamespace(num=2))
    namespace(num=4)
    """

    pass



"""
The idea behind **cached-path** is to provide a unified, simple, extendable interface for accessing
both local and remote files.
This can be used behind other APIs that need to access files agnostic to where they are located.

For remote files, **cached-path** supports several different schemes out-of-the-box in addition
``http`` and ``https``, including ``s3`` for AWS S3, ``gs`` for Google Cloud Storage,
and ``hf`` for HuggingFace Hub. See :func:`cached_path.cached_path()` for more details.

You can also extend **cached-path** to support other schemes with :func:`add_scheme_client()`.
"""

# from ._cached_path import cached_path
# from .common import get_cache_dir, set_cache_dir
# from .progress import get_download_progress
# from .schemes import SchemeClient, add_scheme_client
# from .util import (
#     check_tarfile,
#     filename_to_url,
#     find_latest_cached,
#     is_url_or_existing_file,
#     resource_to_filename,
# )



import logging
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def eval_classification(data=None, **args):
    args = eKonf.to_config(args)
    _eval_ = args._eval_
    labels = args.labels
    if labels:
        labels = eKonf.ensure_list(labels)
    eval_metrics = args[eKonf.Keys.FUNC]
    if data is None:
        data = eKonf.load_data(**args.path.data)
    if data is None:
        log.warning(f"No data to evaluate on given path: {args.path.data}")
        return
    eval_metrics = eKonf.partial(eval_metrics)
    data = data.dropna(subset=[_eval_.actual, _eval_.predicted])
    cm = eval_metrics(data[_eval_.actual], data[_eval_.predicted], labels=labels)
    plot = args.visualize.plot
    plot.plots[0].display_labels = labels
    eKonf.instantiate(plot, data=cm)



# Copyright (c) 2019 NVIDIA CORPORATION. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from bs4 import BeautifulSoup
import wget


class PubMed:
    def __init__(self, subset, output_dir, force_download=False, **kwargs):
        self.subset = subset
        self.force_download = force_download

        self.output_dir = output_dir + "/" + subset
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir, exist_ok=True)

        self.download_urls = {
            "baseline": "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/",
            "daily_update": "https://ftp.ncbi.nlm.nih.gov/pubmed/updatefiles/",
            "oa_comm": "https://ftp.ncbi.nlm.nih.gov/pub/pmc/oa_bulk/oa_comm/xml/",
            "oa_noncomm": "https://ftp.ncbi.nlm.nih.gov/pub/pmc/oa_bulk/oa_noncomm/xml/",
            "oa_other": "https://ftp.ncbi.nlm.nih.gov/pub/pmc/oa_bulk/oa_other/xml/",
        }
        self.download()

    def download(self):
        print("subset:", self.subset)
        if os.listdir(self.output_dir) and not self.force_download:
            print("Files already downloaded. Skipping.")
            return
        url = self.download_urls[self.subset]
        self.download_files(url)

    def download_files(self, url):
        url = self.download_urls[self.subset]
        output = os.popen("curl " + url).read()
        soup = BeautifulSoup(output, "html.parser")
        links = soup.find_all("a")
        filelist = [link.attrs["href"] for link in links]

        if self.subset in ["oa_comm", "oa_noncomm", "oa_other"]:
            filelist = [file for file in filelist if file.endswith(".tar.gz")]
        elif self.subset == "baseline" or self.subset == "daily_update":
            filelist = [file for file in filelist if file.endswith(".gz")]
        else:
            assert False, "Invalid PubMed dataset/subset specified."

        print(f"Number of files: {len(filelist)}")
        for i, filename in enumerate(filelist):
            filepath = os.path.join(self.output_dir, filename)
            if not os.path.isfile(filepath) or os.path.getsize(filepath) == 0:
                print(f"\nDownloading file {i+1} of {len(filelist)}: {filename}")
                wget.download(url + filename, filepath)
            else:
                print(f"{filename} already downloaded. Skipping.")



from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, Sequence, Tuple

from torch import Tensor

from snorkel.classification.data import DictDataLoader  # noqa: F401

BatchIterator = Iterator[
    Tuple[Tuple[Dict[str, Any], Dict[str, Tensor]], "DictDataLoader"]
]


class Scheduler(ABC):
    """Return batches from all dataloaders according to a specified strategy."""

    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_batches(self, dataloaders: Sequence["DictDataLoader"]) -> BatchIterator:
        """Return batches from dataloaders according to a specified strategy.

        Parameters
        ----------
        dataloaders
            A sequence of dataloaders to get batches from

        Yields
        ------
        (batch, dataloader)
            batch is a tuple of (X_dict, Y_dict) and dataloader is the dataloader
            that that batch came from. That dataloader will not be accessed by the
            model; it is passed primarily so that the model can pull the necessary
            metadata to know what to do with the batch it has been given.
        """
